from typing import Dict, Iterable, List, Union

from nortl.core.protocols import ConditionalAssignmentProto, EngineProto, SelectorAssignmentProto
from nortl.renderer.verilog_utils.utils import VerilogRenderable

from .verilog_utils import ENCODINGS, create_state_var
from .verilog_utils.abstractions import MultiHotEncodedStateRegister, OneHotEncodedStateRegister, StateRegister
from .verilog_utils.process import AlwaysComb, AlwaysFF, VerilogAssignment, VerilogIf, VerilogPrint, VerilogPrintf
from .verilog_utils.structural import VerilogDeclaration, VerilogModule, VerilogPortDeclaration

# FIXME: Make empty blocks being not rendered at all.


class VerilogRenderer:
    """This class transforms the engine data into a verilog code.

    To simplify matters, the verilog_utils folder contains code for rendering individual blocks.
    """

    def __init__(
        self,
        engine: EngineProto,
        include_modules: bool = True,
        clock_gating: bool = False,
        encoding: ENCODINGS = 'binary',
        support_unique_if: bool = False,
    ):
        """Initializes the VerilogRenderer with the engine data and rendering options.

        Args:
            engine: The CoreEngine object representing the finite state machine.
            include_modules: A boolean indicating whether to include module instantiations in the generated Verilog code. Defaults to True.
            clock_gating: A boolean indicating whether to enable clock gating logic in the generated Verilog code. Defaults to False.
            encoding: Encoding to use for the state registers
            support_unique_if: If enabled, will render `unique if` and `priority if` statements. This is not supported by all simulators.
        """
        self.engine = engine
        self.verilog_module = VerilogModule(self.engine.module_name)

        self.codelst: List[str] = []

        self.include_modules = include_modules
        self.clock_gating = clock_gating
        self.encoding = encoding
        self.support_unique_if = support_unique_if

        self.clk_request_signals: List[str] = []

        self.state_regs: Dict[str, Union[StateRegister, OneHotEncodedStateRegister, MultiHotEncodedStateRegister]] = {}

    def clear(self) -> None:
        """Clears the internal code list and resets the Verilog module for a new rendering cycle."""
        self.codelst = []
        self.verilog_module = VerilogModule(self.engine.module_name)
        self.state_regs = {}

    def render(self) -> str:
        """Renders the engine into Verilog code.

        This method orchestrates the creation of the Verilog module, signals, instances, and logic blocks.

        Returns:
            A string containing the complete Verilog code.
        """
        self.clear()

        self.create_interface()
        self.create_state_regs()
        self.create_instances()
        self.create_next_state_logic()
        self.create_output_function()
        self.create_prints()
        self.create_combinationals()
        self.create_state_transition()

        if self.clock_gating:
            self.create_clock_gates()

        ret: List[str] = []
        if self.include_modules:
            ret.extend(m.hdl_code for m in self.engine.modules.values())

        ret.append(self.verilog_module.render())

        return '\n'.join(ret)

    def create_clock_gates(self) -> None:
        """Creates the clock gating logic, including gated clock signals and enables.

        This method adds the necessary signals and connections for implementing clock gating,
        which can reduce power consumption by disabling the clock signal to inactive parts of the engine.
        """
        # Create gated clock signals and clock enables
        self.verilog_module.signals.append(VerilogDeclaration('logic', 'GCLK'))
        self.verilog_module.signals.append(VerilogDeclaration('logic', 'GCLK_enable'))
        self.verilog_module.signals.append(VerilogDeclaration('logic', 'GCLK_enable_latched'))

        clk_requests: List[str] = []

        # Get clock requests from instances
        for name, instance in self.engine.module_instances.items():
            if instance.module.clk_request_port is not None:
                signalname = f'clk_request_{name}'
                self.verilog_module.signals.append(VerilogDeclaration('logic', signalname))
                clk_requests.append(signalname)
            if not instance.module._ignore_clk_rst_connection and instance._clock_gating:
                verilog_inst = self.verilog_module.get_instance(name)
                verilog_inst.add_connection('CLK_I', 'GCLK')

        self.create_clock_enable(clk_requests)

        # Change Clock signal of Always_ff blocks
        for blk in self.verilog_module.functionals:
            if isinstance(blk, AlwaysFF):
                blk.clk = 'GCLK'

        # Instantiate Clock Gate
        clock_gate = VerilogDeclaration('nortl_clock_gate', 'I_CLOCK_GATE')
        clock_gate.add_connection('CLK_I', 'CLK_I')
        clock_gate.add_connection('EN', 'GCLK_enable')
        clock_gate.add_connection('GCLK_O', 'GCLK')

        self.verilog_module.instances.append(clock_gate)

    def create_clock_enable(self, clk_requests: List[str]) -> None:
        """Creates the clock enable signal based on state and clock requests.

        This method generates the logic for controlling the clock enable signal, which is used to
        disable the clock signal to inactive parts of the engine, reducing power consumption.
        """
        # Create clock enable by states
        clk_en_proc = AlwaysComb()
        clk_en_proc.add(VerilogAssignment('GCLK_enable', "1'b1"))

        for worker in self.engine.workers.values():
            self.state_regs[worker.name].new_case('next state')

            for state in worker.states:
                if state.has_metadata('Clock_gating'):  # FIXME: Add Metadata to docs
                    self.state_regs[worker.name].add_case(state.name, VerilogAssignment('GCLK_enable', "1'b0"))

                    for condition, _ in state.transitions:
                        block = VerilogIf(condition)
                        block.true_branch.add(VerilogAssignment('GCLK_enable', "1'b1"))
                        self.state_regs[worker.name].add_case(state.name, block)

                    for assignment in state.assignments:
                        if assignment.unconditional:
                            block = VerilogIf((assignment.signal != assignment.value))
                            block.true_branch.add(VerilogAssignment('GCLK_enable', "1'b1"))
                            self.state_regs[worker.name].add_case(state.name, block)
                        else:
                            # FIXME Selectively enable clock for conditional assignments
                            self.state_regs[worker.name].add_case(state.name, VerilogAssignment('GCLK_enable', "1'b1"))

            clk_en_proc.add(self.state_regs[worker.name].build_case())

        for req in clk_requests:
            block = VerilogIf(req)
            block.true_branch.add(VerilogAssignment('GCLK_enable', "1'b1"))
            clk_en_proc.add(block)

        self.verilog_module.functionals.append(clk_en_proc)

    def create_interface(self) -> None:
        """Creates the Verilog module interface, including parameters, signals, and ports.

        This method defines the input and output signals of the Verilog module, as well as any
        parameters that can be used to customize its behavior.
        """
        # Parameters
        for param in self.engine.parameters.values():
            if param.width is not None:
                self.verilog_module.parameters[f'[{param.width - 1}:0] {param.name}'] = param.default_value
            else:
                self.verilog_module.parameters[param.name] = param.default_value

        # Signals and Ports
        self.verilog_module.ports.append(VerilogPortDeclaration('input logic', 'CLK_I'))
        self.verilog_module.ports.append(VerilogPortDeclaration('input logic', 'RST_ASYNC_I'))

        for name, signal in self.engine.signals.items():
            if signal.type == 'local':
                verilog_signal = VerilogDeclaration(signal.data_type, name, width=signal.width)
                self.verilog_module.signals.append(verilog_signal)
            else:
                if signal.data_type in ['reg', 'wire', 'logic']:
                    port_declaration = VerilogPortDeclaration(f'{signal.type} {signal.data_type}', name, signal.width)
                else:
                    port_declaration = VerilogPortDeclaration(signal.data_type, name, width=signal.width)
                self.verilog_module.ports.append(port_declaration)

    def create_state_regs(self) -> None:
        """Creates the Verilog state enumeration.

        This method defines the enumeration type for the engine states, which is used to represent
        the current state of the engine in the Verilog code.
        """

        for name, worker in self.engine.workers.items():
            self.state_regs[name] = create_state_var(worker, self.encoding)
            for state in worker.states:
                self.state_regs[name].add_state(state.name)

        for state_reg in self.state_regs.values():
            self.verilog_module.signals.append(state_reg.declare())

    def create_instances(self) -> None:
        """Creates the Verilog module instances.

        This method instantiates the submodules of the engine, connecting their ports to the
        appropriate signals in the top-level module.
        """
        for instance in self.engine.module_instances.values():
            decl = VerilogDeclaration(instance.module.name, instance.name)

            # Parameters
            for name, value in instance.module.parameters.items():
                if name in instance.parameter_overrides:
                    decl.add_parameter(name, instance.parameter_overrides[name].render())
                else:
                    decl.add_parameter(name, value)

            # Connections
            if not instance.module._ignore_clk_rst_connection:
                decl.add_connection('CLK_I', 'CLK_I')
                decl.add_connection('RST_ASYNC_I', 'RST_ASYNC_I')

            for port_name, signal in instance.port_connections.items():
                decl.add_connection(port_name, signal.render())

            self.verilog_module.instances.append(decl)

    def create_next_state_logic(self) -> None:
        """Creates the Verilog next state logic.

        This method generates the logic for determining the next state of the engine, based on the
        current state and the input signals.
        """
        next_state_func = AlwaysComb()

        for worker in self.engine.workers.values():
            self.state_regs[worker.name].new_case('current state')

            next_state_func.add(self.state_regs[worker.name].state_transition(None))

            for state in worker.states:
                for condition, next_state in state.transitions:
                    if condition.render() == "1'h1":
                        self.state_regs[worker.name].add_case(state.name, self.state_regs[worker.name].state_transition(next_state.name))
                    elif condition.render() == "1'h0":
                        pass
                    else:
                        item = VerilogIf(condition)
                        item.true_branch.add(self.state_regs[worker.name].state_transition(next_state.name))
                        self.state_regs[worker.name].add_case(state.name, item)

            next_state_func.add(self.state_regs[worker.name].build_case())

        self.verilog_module.functionals.append(next_state_func)

    def create_output_function(self) -> None:  # noqa: C901
        """Creates the Verilog output function.

        This method generates the logic for updating the output signals of the engine, based on the
        current state and the input signals.
        """
        output_func = AlwaysFF('CLK_I', 'RST_ASYNC_I')

        for assignment in self.engine.main_worker.reset_state.assignments:
            if not assignment.unconditional:
                raise RuntimeError('Conditional assignments are not allowed in the reset state.')
            output_func.add_reset(VerilogAssignment(assignment.signal, assignment.value))

        for signal in self.engine.signals.values():
            if signal.pulsing:
                output_func.add(VerilogAssignment(signal, '0'))

        for worker in self.engine.workers.values():
            self.state_regs[worker.name].new_case('next state')

            for state in worker.states:
                if len(state.assignments) > 0:
                    for assignment in state.assignments:
                        if assignment.unconditional:
                            self.state_regs[worker.name].add_case(state.name, VerilogAssignment(assignment.signal, assignment.value))
                        else:
                            for block in self._extract_conditional_assignment(assignment, support_unique_if=self.support_unique_if):
                                self.state_regs[worker.name].add_case(state.name, block)

            output_func.add(self.state_regs[worker.name].build_case())

        self.verilog_module.functionals.append(output_func)

    @classmethod
    def _extract_conditional_assignment(
        cls, assignment: Union[ConditionalAssignmentProto, SelectorAssignmentProto], support_unique_if: bool = False
    ) -> Iterable[VerilogRenderable]:
        for i, (condition, value) in enumerate(assignment.cases):
            if i == 0:
                if support_unique_if:
                    block = VerilogIf(condition, keyword='priority if' if assignment.priority else 'unique if')
                else:
                    block = VerilogIf(condition, keyword='if')
            else:
                block = VerilogIf(condition, keyword='else if')

            if hasattr(value, 'cases'):
                # Value is an selector assignment itself, turn it into nested If, Else If, etc.
                for nested_block in cls._extract_conditional_assignment(value, support_unique_if=support_unique_if):  # type: ignore[arg-type]
                    block.true_branch.add(nested_block)
            else:
                block.true_branch.add(VerilogAssignment(assignment.signal, value))

            # Close off the list of cases with a default
            if i == (len(assignment.cases) - 1) and assignment.default is not None:
                if hasattr(assignment.default, 'cases'):
                    for nested_block in cls._extract_conditional_assignment(assignment.default, support_unique_if=support_unique_if):  # type: ignore[arg-type]
                        block.false_branch.add(nested_block)
                else:
                    block.false_branch.add(VerilogAssignment(assignment.signal, assignment.default))
            yield block

    def create_prints(self) -> None:
        """Creates the print statements that have been put into each state."""
        output_func = AlwaysFF('CLK_I', 'RST_ASYNC_I')

        for worker in self.engine.workers.values():
            self.state_regs[worker.name].new_case('next state')

            for state in worker.states:
                for fname, item in state._printfs.items():
                    for line, val in item:
                        self.state_regs[worker.name].add_case(state.name, VerilogPrintf(fname, line, *[v.render() for v in val]))

                for line, val in state._prints:
                    self.state_regs[worker.name].add_case(state.name, VerilogPrint(line, *[v.render() for v in val]))

            output_func.add(self.state_regs[worker.name].build_case())

        self.verilog_module.functionals.append(output_func)

    def create_combinationals(self) -> None:
        """Creates the Verilog combinational logic.

        This method generates the combinational logic for the engine, based on the input signals.
        """
        combinationals = AlwaysComb()

        for signal, value in self.engine.combinationals:
            combinationals.add(VerilogAssignment(signal, value))

        self.verilog_module.functionals.append(combinationals)

    def create_state_transition(self) -> None:
        """Creates the Verilog state transition logic.

        This method generates the logic for transitioning the engine from one state to another.
        """
        for state_reg in self.state_regs.values():
            self.verilog_module.functionals.append(state_reg.build_state_transition())



"""Tests correct handling of input and output slots, as well as signal-expiration."""

import re
from typing import Iterable, Iterator, Protocol

import pytest

from nortl import Engine, Segment
from nortl.core.protocols import ScratchSignalProto, SignalProto


class TestExecutor(Protocol):
    __test__: bool = False

    def __call__(self, engine: Engine, clock_gating: bool = False) -> str: ...


def sync_between[T](engine: Engine, it: Iterable[T]) -> Iterator[T]:
    it = iter(it)
    yield next(it)
    while (elem := next(it, None)) is not None:
        engine.sync()
        yield elem


@Segment.with_input_slots(value=8).with_output_slots(8)
def incr(engine: Engine, value: SignalProto) -> ScratchSignalProto:
    engine.current_state.name = 'start_state'
    result = engine.define_scratch(8)
    engine.set(result, value)
    engine.sync()

    engine.set(result, result + 1)
    engine.current_state.name = 'end_state'
    return result


def test_copy(execute_test: TestExecutor) -> None:
    """Test copying values in and out of the segment."""
    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    output = engine.define_output('OUT', 8, reset_value=0)
    engine.sync()
    assert len(engine.states['main']) == 2

    for i in range(4):
        engine.current_state.name = f'call_state{i}'
        output_plus_one = incr(engine, output)
        engine.current_state.name = f'return_state{i}'
        engine.set(output, output_plus_one)

        # Each call takes 1 state (return/call), because return and call overlap
        # The first call takes two more states (start and end state inside the segment)
        assert len(engine.states['main']) == 5 + i

    # Execute test
    result = execute_test(engine)
    print(result)
    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(5)]

    assert res == expected


class MyEngine(Engine):
    def __init__(self, module_name: str, reset_state_name: str = 'IDLE') -> None:
        super().__init__(module_name, reset_state_name)

        # Define a result pointer as a fixed resource for the segment
        # Note that this signal must not be accessed directly outside of the segment, or else the reference-invalidation
        # feature will not apply.
        self.result_pointer = self.define_local('result_pointer', 8, reset_value=0)

    # This function always returns the same signal
    @Segment.with_input_slots(value=8)
    def incr_with_ptr(self, value: SignalProto) -> SignalProto:
        self.current_state.name = 'start_state'
        self.set(self.result_pointer, value)
        self.sync()

        self.set(self.result_pointer, self.result_pointer + 1)
        self.current_state.name = 'end_state'

        return self.result_pointer


def test_signal_reference(execute_test: TestExecutor) -> None:
    """Test Weak Reference invalidation of returned signals.

    When a segment returns a regular signal, and doesn't have an (unnecessary) output slot for it, the signal is wrapped in a WeakReference modifier.
    This reference is invalidated/expired when the segment is called again, because then the underlying signal will already contain the next result.

    Warning: this safety feature can be bypassed, if the signal is not explicitely returned from the segment.
    """
    engine = MyEngine('my_engine')
    _ = engine.define_input('IN')
    output = engine.define_output('OUT', 8, reset_value=0)
    engine.sync()

    previous_output_plus_one: SignalProto = None  # type: ignore
    for i in range(4):
        engine.current_state.name = f'call_state{i}'

        output_plus_one = engine.incr_with_ptr(output)
        engine.current_state.name = f'return_state{i}'
        engine.set(output, output_plus_one)
        engine.sync()

        # output_plus_one will contain a WeakReference modifier of the signal
        # It gets invalidated once the segment is called again
        if i > 0:
            with pytest.raises(RuntimeError):
                engine.set(output, previous_output_plus_one)

        previous_output_plus_one = output_plus_one

        # Each call now takes 2 states (call, return), because no copy-out state is needed
        # The first call takes two more states
        assert len(engine.states['main']) == 6 + 2 * i

    # Execute test
    result = execute_test(engine)
    print(result)
    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(5)]

    assert res == expected



"""Tests correct functionality of the segment decorator."""

import re
from typing import Iterable, Iterator, Protocol

from nortl import Engine, Segment
from nortl.core.protocols import SignalProto


class TestExecutor(Protocol):
    __test__: bool = False

    def __call__(self, engine: Engine, clock_gating: bool = False) -> str: ...


def sync_between[T](engine: Engine, it: Iterable[T]) -> Iterator[T]:
    it = iter(it)
    yield next(it)
    while (elem := next(it, None)) is not None:
        engine.sync()
        yield elem


# Test different styles of decoration
# Segments can be added to functions, or methods on subclasses of an noRTL engine, or on methods on helper classes
@Segment
def count_to_four(engine: Engine, counter: SignalProto) -> None:
    for _ in sync_between(engine, range(4)):
        engine.set(counter, counter + 1)


def test_function_segment(execute_test: TestExecutor) -> None:
    """Test calling the segment function."""
    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    output = engine.define_output('OUT', 8, reset_value=0)
    engine.sync()
    assert len(engine.states['main']) == 2

    # Call the segment (output will be at 4)
    # This will create 5 new states (4 within the segment, and one return state)
    count_to_four(engine, output)
    assert len(engine.states['main']) == 7

    # Call the segment again
    # This will only create 1 new state for the new, second return state!
    count_to_four(engine, output)
    assert len(engine.states['main']) == 8

    # Execute test
    result = execute_test(engine)
    print(result)
    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(9)]

    assert res == expected


class MyEngine(Engine):
    @Segment
    def count_to_four(self, counter: SignalProto) -> None:
        for _ in sync_between(self, range(4)):
            self.set(counter, counter + 1)


def test_method_segment(execute_test: TestExecutor) -> None:
    """Test adding a segment to a class method."""
    engine = MyEngine('my_engine')
    _ = engine.define_input('IN')
    output = engine.define_output('OUT', 8, reset_value=0)
    engine.sync()

    # Call the segment
    engine.count_to_four(output)
    assert len(engine.states['main']) == 7

    # Call the segment again
    engine.count_to_four(output)
    assert len(engine.states['main']) == 8

    # Execute test
    result = execute_test(engine)
    print(result)
    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(9)]

    assert res == expected


class MyHelper:
    def __init__(self, engine: Engine):
        self.engine = engine  # the engine must always be stored in a attribute engine, for the segment to work

    @Segment
    def count_to_four(self, counter: SignalProto) -> None:
        for _ in sync_between(self.engine, range(4)):
            self.engine.set(counter, counter + 1)


def test_helper_method_segment(execute_test: TestExecutor) -> None:
    """Test adding a segment to a helper class method."""
    engine = MyEngine('my_engine')
    _ = engine.define_input('IN')
    output = engine.define_output('OUT', 8, reset_value=0)
    engine.sync()

    helper = MyHelper(engine)

    # Call the segment
    helper.count_to_four(output)
    assert len(engine.states['main']) == 7

    # Call the segment again
    helper.count_to_four(output)
    assert len(engine.states['main']) == 8

    # Execute test
    result = execute_test(engine)
    print(result)
    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(9)]

    assert res == expected


# Test inlining
def test_function_inline(execute_test: TestExecutor) -> None:
    """Test calling the segment function as an inline.

    Effectively, this treats the decorated function as if it wouldn't have the decorator.
    """
    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    output = engine.define_output('OUT', 8, reset_value=0)
    engine.sync()

    # Call the segment as an inline
    # Inlining will save two states for the entry and exit.
    # However, an additional sync is still required here, that is merged into the exit state of the call by the sync_between.
    count_to_four.inline(engine, output)
    engine.sync()
    assert len(engine.states['main']) == 6

    # However, calling the segment again, will add 3 new states (because there are 3x sync() in the function)
    count_to_four.inline(engine, output)
    engine.sync()
    assert len(engine.states['main']) == 10

    # Execute test
    result = execute_test(engine)
    print(result)
    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(9)]

    assert res == expected


def test_method_inline(execute_test: TestExecutor) -> None:
    """Test calling the segment method as an inline."""

    engine = MyEngine('my_engine')
    _ = engine.define_input('IN')
    output = engine.define_output('OUT', 8, reset_value=0)
    engine.sync()

    # Call the segment
    engine.count_to_four.inline(output)
    engine.sync()
    assert len(engine.states['main']) == 6

    # Call the segment again
    engine.count_to_four.inline(output)
    engine.sync()
    assert len(engine.states['main']) == 10

    # Execute test
    result = execute_test(engine)
    print(result)
    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(9)]

    assert res == expected


def test_helper_method_inline(execute_test: TestExecutor) -> None:
    """Test calling the segment method as an inline."""

    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    output = engine.define_output('OUT', 8, reset_value=0)
    engine.sync()

    helper = MyHelper(engine)

    # Call the segment
    helper.count_to_four.inline(output)
    engine.sync()
    assert len(engine.states['main']) == 6

    # Call the segment again
    helper.count_to_four.inline(output)
    engine.sync()
    assert len(engine.states['main']) == 10

    # Execute test
    result = execute_test(engine)
    print(result)
    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(9)]

    assert res == expected



from pathlib import Path

from nortl import Engine
from nortl.renderer import ScratchpadVisualizationRenderer


def test_simple_rendering() -> None:
    # Some whatever engine
    engine = Engine('my_engine')
    engine.tracer.upper_boundary = str(Path(__file__))

    _ = engine.define_input('IN')
    engine.sync()
    local_cnt = engine.define_scratch(4)
    engine.set(local_cnt, 1)
    dummy_before = engine.define_local('test', 2, 0)
    engine.set(dummy_before, local_cnt)

    with engine.fork('proc2') as proc2:
        with engine.fork('Proc1') as proc1:
            dummy = engine.define_scratch(2)
            engine.set(dummy, local_cnt)
            engine.sync()
            engine.set(local_cnt, 2)

        proc1.cancel()
        engine.set(local_cnt, 2)
        engine.set(dummy_before, local_cnt)

    proc2.cancel()
    dummy = engine.define_scratch(2)
    engine.set(dummy, local_cnt)
    engine.sync()

    renderer = ScratchpadVisualizationRenderer(engine)

    with open((Path(__file__).parent / 'artifacts') / 'out.md', 'w') as file:
        file.write(renderer.render())


def test_simple_rendering_2() -> None:
    # Some whatever engine
    engine = Engine('my_engine')
    engine.tracer.upper_boundary = str(Path(__file__))

    _ = engine.define_input('IN')
    engine.sync()
    local_cnt = engine.define_scratch(4)
    engine.set(local_cnt, 1)
    _ = engine.define_scratch(2)
    engine.sync()
    engine.sync()
    local_cnt.release()
    engine.sync()

    _ = engine.define_scratch(2)

    engine.sync()

    renderer = ScratchpadVisualizationRenderer(engine)

    with open((Path(__file__).parent / 'artifacts') / 'out2.md', 'w') as file:
        file.write(renderer.render())



from contextlib import contextmanager
from inspect import FrameInfo
from typing import (
    Any,
    ClassVar,
    ContextManager,
    Dict,
    Generic,
    Iterator,
    List,
    Literal,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from typing_extensions import Self

from nortl.utils.type_aliases import IntSlice

__all__ = [
    'AnySignal',
    'Operand',
    'PermanentSignal',
    'RenderableSelector',
    'Selector',
]

# Operand types
Operand = Union['Renderable', int, bool]
Selector = Mapping[Union['Renderable', Literal['default']], Union[Operand, 'Selector']]
RenderableSelector = Mapping[Union['Renderable', Literal['default']], Union['Renderable', 'RenderableSelector']]

# Signal types
PermanentSignal = Union['SignalProto', 'SignalSliceProto']
AnySignal = Union['SignalProto', 'SignalSliceProto', 'ScratchSignalProto']

# Assignment types
AnyAssignment = Union['AssignmentProto', 'ConditionalAssignmentProto', 'SelectorAssignmentProto']

SIGNAL_TYPES = Literal['input', 'output', 'interface', 'internal', 'local']
EVENT_TYPES = Literal['edge', 'delay', 'sync']
BIT_ORDER = Literal['H:L', 'L:H']

# Access Check Definitions
SIGNAL_ACCESS_CHECKS = Literal['exclusive_read', 'exclusive_write', 'identical_rw']
ACCESS_CHECKS = SIGNAL_ACCESS_CHECKS

T_Signal = TypeVar('T_Signal', 'SignalProto', 'SignalSliceProto', covariant=True)


class NamedEntityProto(Protocol):
    @property
    def name(self) -> str: ...

    def get_metadata(self, key: str, default: Any = None) -> Any: ...

    def set_metadata(self, key: str, value: Any) -> None: ...

    def has_metadata(self, key: str) -> bool: ...


class OperationTraitProto(Protocol):
    """Trait for signals, constants or statements that allow construction of arithmetic and logic operations."""

    @property
    def is_primitive(self) -> bool: ...

    @property
    def is_constant(self) -> bool: ...

    @property
    def operand_width(self) -> Optional[int]: ...

    def read_access(self, ignore: Set[ACCESS_CHECKS] = ...) -> None: ...

    def render(self, target: Optional[str] = None) -> str: ...

    def __hash__(self) -> int: ...

    # Arithemtic Operations
    def __add__(self, value: Operand, /) -> 'Renderable': ...

    def __sub__(self, value: Operand, /) -> 'Renderable': ...

    def __mul__(self, value: Operand, /) -> 'Renderable': ...

    def __truediv__(self, value: Operand, /) -> 'Renderable': ...

    def __mod__(self, value: Operand, /) -> 'Renderable': ...

    # Arithmetic Operations (Right-Side)
    def __radd__(self, value: Operand, /) -> 'Renderable': ...

    def __rsub__(self, value: Operand, /) -> 'Renderable': ...

    def __rmul__(self, value: Operand, /) -> 'Renderable': ...

    def __rtruediv__(self, value: Operand, /) -> 'Renderable': ...

    def __rmod__(self, value: Operand, /) -> 'Renderable': ...

    # Logic Operations
    def __and__(self, value: Operand, /) -> 'Renderable': ...

    def __or__(self, value: Operand, /) -> 'Renderable': ...

    def __xor__(self, value: Operand, /) -> 'Renderable': ...

    def __lshift__(self, value: Operand, /) -> 'Renderable': ...

    def __rshift__(self, value: Operand, /) -> 'Renderable': ...

    # Logic Operations (Right Side)
    def __rand__(self, value: Operand, /) -> 'Renderable': ...

    def __ror__(self, value: Operand, /) -> 'Renderable': ...

    def __rxor__(self, value: Operand, /) -> 'Renderable': ...

    def __rlshift__(self, value: Operand, /) -> 'Renderable': ...

    def __rrshift__(self, value: Operand, /) -> 'Renderable': ...

    # Misc.
    def __neg__(self) -> 'Renderable': ...

    def __pos__(self) -> 'Renderable': ...

    # Inversion
    def __invert__(self) -> 'Renderable': ...

    # Comparison
    def __eq__(self, value: Operand, /) -> 'Renderable': ...  # type: ignore[override]

    def __ne__(self, value: Operand, /) -> 'Renderable': ...  # type: ignore[override]

    def __lt__(self, value: Operand, /) -> 'Renderable': ...

    def __le__(self, value: Operand, /) -> 'Renderable': ...

    def __gt__(self, value: Operand, /) -> 'Renderable': ...

    def __ge__(self, value: Operand, /) -> 'Renderable': ...

    # Bit Slicing
    def __getitem__(self, index: Union[int, slice]) -> 'OperationTraitProto': ...


class OperationProto(OperationTraitProto, Protocol):
    """Wrapper object that represents an operation."""

    @property
    def operands(self) -> Sequence['Renderable']: ...


class Renderable(OperationTraitProto, Protocol):
    """Renderable signal or operation that can be used as an condition."""


class AssignmentTarget(OperationTraitProto, Protocol):
    """Object that can be used as assignment target."""

    @property
    def name(self) -> str: ...

    @property
    def engine(self) -> 'EngineProto': ...

    def write_access(self, ignore: Set[ACCESS_CHECKS] = ...) -> None: ...

    def overlaps_with(self, other: 'AssignmentTarget') -> Union[bool, Literal['partial']]: ...


class ConstProto(Renderable, Protocol):
    pass


class VarProto(ConstProto, Protocol):
    def update(self, value: Union[int, bool, str]) -> None: ...


class ModuleInstanceProto(Protocol):
    _clock_gating: bool

    @property
    def module(self) -> 'ModuleProto': ...

    @property
    def name(self) -> str: ...

    @property
    def port_connections(self) -> Dict[str, 'PermanentSignal']: ...

    @property
    def parameter_overrides(self) -> Dict[str, Union[ConstProto, 'ParameterProto', Renderable]]: ...

    def connect_port(self, port_name: str, signal: 'PermanentSignal') -> None: ...

    def get_connected_signal(self, port_name: str) -> 'PermanentSignal': ...

    def override_parameter(self, name: str, value: Union[int, 'ParameterProto', Renderable]) -> None: ...


# Assignments
class AssignmentProto(Protocol):
    signal: AssignmentTarget
    value: Renderable

    @property
    def unconditional(self) -> Literal[True]: ...


class ConditionalAssignmentProto(Protocol):
    signal: AssignmentTarget
    value: Renderable
    condition: Renderable

    @property
    def priority(self) -> Literal[False]: ...

    @property
    def unconditional(self) -> Literal[False]: ...

    @property
    def cases(self) -> Tuple[Tuple[Renderable, Renderable]]: ...

    @property
    def default(self) -> Optional[Renderable]: ...

    def flatten_cases(self) -> Sequence[Tuple[Renderable, Renderable]]: ...


class SelectorAssignmentProto(Protocol):
    signal: AssignmentTarget
    selector: RenderableSelector

    @property
    def priority(self) -> bool: ...

    @property
    def unconditional(self) -> Literal[False]: ...

    @property
    def cases(self) -> Sequence[Tuple[Renderable, Union[Renderable, 'SelectorAssignmentProto']]]: ...

    @property
    def default(self) -> Optional[Union[Renderable, 'SelectorAssignmentProto']]: ...

    def flatten_cases(self) -> Sequence[Tuple[Renderable, Renderable]]: ...


class StateProto(NamedEntityProto, Protocol):
    _prints: List[Tuple[str, Tuple[Renderable, ...]]]
    _printfs: Dict[str, List[Tuple[str, Tuple[Renderable, ...]]]]

    active_scratch_signals: List['ScratchSignalProto']

    @property
    def engine(self) -> 'EngineProto': ...

    @property
    def worker(self) -> 'WorkerProto': ...

    # Assignment Management
    @property
    def allow_assignments(self) -> bool: ...

    @property
    def assignments(self) -> Sequence[AnyAssignment]: ...

    def add_assignment(self, signal: AssignmentTarget, value: Renderable, condition: Optional[Renderable] = None) -> None: ...

    def add_selector_assignment(self, signal: AssignmentTarget, selector: RenderableSelector, allow_short_circuit: bool = False) -> None: ...

    def get_assignments(self, signal: AssignmentTarget) -> Iterator[AnyAssignment]: ...

    # Transition Management
    @property
    def transitions(self) -> Sequence[Tuple[Renderable, Self]]: ...

    def _add_transition(self, condition: Renderable, state: Self) -> None: ...

    def _restrict_transition(self, state: Self) -> None: ...

    def _lock_transitions(self) -> None: ...

    # Misc.
    def render(self, target: Optional[str] = None) -> str: ...

    def print(self, line: str, *args: Renderable) -> None: ...

    def printf(self, fname: str, line: str, *args: Renderable) -> None: ...

    @property
    def signature(self) -> str: ...


class ScratchManagerProto(Protocol):
    @property
    def engine(self) -> 'EngineProto': ...

    # Scratch signal management
    def create_signal(self, width: int) -> 'ScratchSignalProto': ...

    # Context managememt
    def enter_context(self) -> None: ...

    def exit_context(self) -> None: ...

    def force_release_signals_by_thread(self, thread: 'ThreadProto') -> None: ...

    def free_accesses_from_thread(self, thread: 'ThreadProto') -> None: ...

    # Memory zone managment
    @property
    def zones(self) -> Sequence['MemoryZoneProto']: ...

    @property
    def main_zone(self) -> 'MemoryZoneProto': ...

    @property
    def active_zone(self) -> 'MemoryZoneProto': ...

    @property
    def suspended_zones(self) -> Sequence['MemoryZoneProto']: ...

    def create_zone(self, name: Optional[str] = None) -> 'MemoryZoneProto': ...


class MemoryZoneProto(ContextManager['MemoryViewProto'], Protocol):
    @property
    def manager(self) -> ScratchManagerProto: ...

    @property
    def name(self) -> str: ...

    @property
    def id(self) -> int: ...

    # Scratch pad
    width: int

    @property
    def scratchpad(self) -> 'SignalProto': ...

    @property
    def scratch_map(self) -> Sequence[bool]: ...

    def alloc(self, width: int) -> Optional[int]: ...

    # Scratch signal management
    @property
    def scratch_signals(self) -> Sequence['ScratchSignalProto']: ...

    def create_signal(self, width: int) -> 'ScratchSignalProto': ...

    # Activation and view management
    @property
    def active(self) -> bool: ...

    @property
    def suspended(self) -> bool: ...

    @property
    def active_view(self) -> 'MemoryViewProto': ...

    @property
    def views(self) -> Sequence['MemoryViewProto']: ...

    @contextmanager
    def recover(self, view: 'MemoryViewProto') -> Iterator[None]: ...


class MemoryViewProto(Protocol):
    @property
    def zone(self) -> MemoryZoneProto: ...

    @property
    def scratch_signals(self) -> Sequence['ScratchSignalProto']: ...

    def create_signal(self, width: int) -> 'ScratchSignalProto': ...

    def _register_signal(self, signal: 'ScratchSignalProto') -> None: ...


class SignalManagerProto(Protocol):
    @property
    def signals(self) -> Mapping[str, 'SignalProto']: ...

    @property
    def combinationals(self) -> Sequence[Tuple['SignalProto', Renderable]]: ...

    def get_signal(self, name: str) -> 'SignalProto': ...

    def create_signal(
        self,
        type: SIGNAL_TYPES,
        name: str,
        width: Union[int, 'ParameterProto', Renderable] = 1,
        data_type: str = 'logic',
        is_synchronized: bool = False,
        pulsing: bool = False,
    ) -> 'SignalProto': ...

    def free_accesses_from_thread(self, thread: 'ThreadProto') -> None: ...


# Protocol mixin, used by Engine and Worker
class _StateManager(Protocol):
    @property
    def state_names(self) -> Set[str]: ...

    def create_state(self, name: Optional[str] = None, allow_assignments: bool = True, metadata: Dict[str, Any] = {}) -> 'StateProto': ...

    @property
    def current_state(self) -> 'StateProto': ...

    @current_state.setter
    def current_state(self, state: 'StateProto') -> None: ...

    @property
    def next_state(self) -> 'StateProto': ...

    @property
    def reset_state(self) -> 'StateProto': ...


class EngineProto(_StateManager, Protocol):
    MAIN_WORKER_NAME: ClassVar[str]
    MAIN_THREAD_NAME: ClassVar[str]

    module_name: str
    state_metadata_template: Dict[str, Any]

    # Tracing
    @property
    def tracer(self) -> 'TracerProto': ...

    # State manegement (+_StateManager)
    @property
    def states(self) -> Mapping[str, Sequence[StateProto]]: ...

    # Worker Managment
    @property
    def workers(self) -> Mapping[str, 'WorkerProto']: ...

    def create_worker(self, name: Optional[str] = None) -> 'WorkerProto': ...

    @property
    def current_worker(self) -> 'WorkerProto': ...

    @current_worker.setter
    def current_worker(self, worker: 'WorkerProto') -> None: ...

    @property
    def main_worker(self) -> 'WorkerProto': ...

    # Thread Managment
    @property
    def current_thread(self) -> 'ThreadProto': ...

    @property
    def main_thread(self) -> 'ThreadProto': ...

    # Signal Managment
    @property
    def signal_manager(self) -> SignalManagerProto: ...

    @property
    def scratch_manager(self) -> ScratchManagerProto: ...

    @property
    def signals(self) -> Mapping[str, 'SignalProto']: ...

    @property
    def combinationals(self) -> Sequence[Tuple['SignalProto', Renderable]]: ...

    def define_input(
        self, name: str, width: Union[int, 'ParameterProto', Renderable] = 1, data_type: str = 'logic', is_synchronized: bool = False
    ) -> 'SignalProto': ...

    def define_output(
        self,
        name: str,
        width: Union[int, 'ParameterProto', Renderable] = 1,
        reset_value: int = 0,
        data_type: str = 'logic',
        value: Union[Renderable, None] = None,
    ) -> 'SignalProto': ...

    def define_local(
        self,
        name: str,
        width: Union[int, 'ParameterProto', Renderable] = 1,
        reset_value: int | None = None,
        data_type: str = 'logic',
        pulsing: bool = False,
        value: Union[Renderable, None] = None,
    ) -> 'SignalProto': ...

    def define_scratch(self, width: int) -> 'ScratchSignalProto': ...

    # Parameter Managment
    @property
    def parameters(self) -> Mapping[str, 'ParameterProto']: ...

    def define_parameter(self, name: str, default_value: int = 0, width: Optional[int] = None) -> 'ParameterProto': ...

    # Setting outputs
    def set(self, signal: AssignmentTarget, level: Union[Renderable, int, bool]) -> None: ...

    def set_when(self, signal: AssignmentTarget, selector: Selector, allow_short_circuit: bool = False) -> None: ...

    def set_once(self, signal: AssignmentTarget, level: Union[Renderable, int, bool], reset_level: Union[Renderable, int, bool] = False) -> None: ...

    def sync(self) -> None: ...

    def wait_for(self, condition: Renderable) -> None: ...

    def jump_if(self, condition: Renderable, true_state: StateProto, false_state: Optional[StateProto] = None) -> None: ...

    # Debugging Prints

    def print(self, line: str, *args: Renderable) -> None: ...

    def printf(self, fname: str, line: str, *args: Renderable) -> None: ...

    # Module Managment
    @property
    def modules(self) -> Mapping[str, 'ModuleProto']: ...

    @property
    def module_instances(self) -> Mapping[str, ModuleInstanceProto]: ...

    def define_module(self, name: str) -> 'ModuleProto': ...

    def add_module(self, module: 'ModuleProto') -> None: ...

    def create_module_instance(self, module_name: str, instance_name: str, clock_gating: bool = False) -> 'ModuleInstanceProto': ...

    def connect_module_port(self, instance_name: str, port_name: str, signal: 'PermanentSignal') -> None: ...

    def override_module_parameter(self, instance_name: str, parameter_name: str, value: Union[int, 'ParameterProto', Renderable]) -> None: ...


class ModuleProto(Protocol):
    _ignore_clk_rst_connection: bool

    @property
    def name(self) -> str: ...

    @property
    def ports(self) -> List[str]: ...
    @property
    def parameters(self) -> Dict[str, int]: ...

    @property
    def hdl_code(self) -> str: ...

    def add_port(self, port_name: str) -> None: ...

    def has_port(self, port_name: str) -> bool: ...

    def add_parameter(self, name: str, value: int) -> None: ...

    def set_clk_request(self, port_name: str) -> None: ...

    @property
    def clk_request_port(self) -> Optional[str]: ...


# Access checks
class StaticAccessProto(Protocol):
    thread: 'ThreadProto'
    active: bool

    def disable(self) -> None: ...


class StaticAccessCheckerProto(Protocol):
    @property
    def reading_thread_names(self) -> Set[str]: ...

    @property
    def writing_thread_names(self) -> Set[str]: ...

    @property
    def all_checks_enabled(self) -> bool: ...

    def disable_check(self, check: SIGNAL_ACCESS_CHECKS) -> None: ...

    def check(self, ignore: Set[ACCESS_CHECKS] = set()) -> None: ...


# Signals
class _BaseSignalProto(AssignmentTarget, Protocol):
    @property
    def _is_synchronized(self) -> bool: ...

    @property
    def type(self) -> SIGNAL_TYPES: ...

    @property
    def pulsing(self) -> bool: ...

    @property
    def escaped_name(self) -> str: ...

    @property
    def width(self) -> Union[int, 'ParameterProto', Renderable]: ...

    @property
    def data_type(self) -> str: ...

    # Access Control
    @property
    def read_accesses(self) -> Set[StaticAccessProto]: ...

    @property
    def write_accesses(self) -> Set[StaticAccessProto]: ...

    @property
    def last_read_access_thread(self) -> Optional['ThreadProto']: ...

    @last_read_access_thread.setter
    def last_read_access_thread(self, value: 'ThreadProto') -> None: ...

    @property
    def last_write_access_thread(self) -> Optional['ThreadProto']: ...

    @last_write_access_thread.setter
    def last_write_access_thread(self, value: 'ThreadProto') -> None: ...

    @property
    def access_checker(self) -> StaticAccessCheckerProto: ...

    def free_access_from_thread(self, thread: 'ThreadProto') -> None: ...


class _EventSourceSignalProto(Generic[T_Signal], Protocol):
    def rising(self) -> T_Signal: ...

    def falling(self) -> T_Signal: ...

    def delayed(self, cycles: Union[int, 'ParameterProto'] = 1) -> T_Signal: ...

    def synchronized(self) -> T_Signal: ...


class _BaseSliceProto(_BaseSignalProto, Protocol):
    @property
    def base_signal(self) -> 'SignalProto': ...

    @property
    def index(self) -> Union[int, IntSlice]: ...

    def __getitem__(self, index: Union[int, IntSlice]) -> Self: ...


class SignalProto(_EventSourceSignalProto['SignalProto'], _BaseSignalProto, Protocol):
    def __getitem__(self, index: Union[int, IntSlice]) -> 'SignalSliceProto': ...


class SignalSliceProto(_BaseSliceProto, _EventSourceSignalProto['SignalSliceProto'], Protocol):
    def as_scratch_signal(self, zone: Optional[MemoryZoneProto] = None) -> 'ScratchSignalProto': ...


class ScratchSignalProto(_BaseSliceProto, ContextManager['ScratchSignalProto'], Protocol):
    creator_frames: Sequence[FrameInfo]

    @property
    def owner(self) -> 'ThreadProto': ...

    @property
    def zone(self) -> MemoryZoneProto: ...

    def enter_context(self) -> None: ...

    def exit_context(self) -> None: ...

    @property
    def released(self) -> bool: ...

    def release(self, force: bool = False) -> None: ...

    def reclaim(self) -> None: ...

    def states_disjoint(self, other: Self) -> bool: ...

    def call_stack_similarity(self, other: Self) -> int: ...


# Parameters
class ParameterProto(Renderable, Protocol):
    @property
    def engine(self) -> EngineProto: ...

    @property
    def name(self) -> str: ...

    @property
    def default_value(self) -> int: ...

    @property
    def width(self) -> Optional[int]: ...


# Parallel processing
class WorkerProto(_StateManager, NamedEntityProto, Protocol):
    @property
    def engine(self) -> EngineProto: ...

    @property
    def is_main_worker(self) -> bool: ...

    def create_scoped_name(self, name: str) -> str: ...

    @property
    def sync_reset(self) -> Renderable: ...

    @sync_reset.setter
    def sync_reset(self, value: Renderable) -> None: ...

    # Control Signals for Threads
    @property
    def reset(self) -> SignalProto: ...

    @property
    def start(self) -> SignalProto: ...

    @property
    def select(self) -> SignalProto: ...

    @property
    def idle(self) -> SignalProto: ...

    # State Management
    @property
    def states(self) -> Sequence[StateProto]: ...

    def leave_foreground(self) -> None: ...

    # Thread Management
    @property
    def threads(self) -> Sequence['ThreadProto']: ...

    def create_thread(self, name: Optional[str] = None) -> 'ThreadProto': ...

    @property
    def working(self) -> bool: ...

    @property
    def current_thread(self) -> 'ThreadProto': ...

    # Misc.
    def raise_reset(self) -> None: ...

    def clear_reset(self) -> None: ...


class ThreadProto(NamedEntityProto, Protocol):
    @property
    def engine(self) -> EngineProto: ...

    @property
    def worker(self) -> WorkerProto: ...

    @property
    def is_main_thread(self) -> bool: ...

    @property
    def active(self) -> bool: ...
    @active.setter
    def active(self, value: bool) -> None: ...

    _spawned_threads: List[Self]
    parent_thread: Optional[Self]

    def join(self) -> None: ...

    @property
    def finished(self) -> Renderable: ...

    @property
    def running(self) -> bool: ...

    def cancel(self) -> None: ...

    def finish(self) -> None: ...

    @property
    def call_stack(self) -> List[Self]: ...


class TracerProto(Protocol):
    def add_metadata(self, target: NamedEntityProto, key: str, profile: bool = False) -> None: ...

    def format_metadata(self, target: NamedEntityProto, key: str) -> str: ...

    @property
    def current_trace(self) -> Sequence[FrameInfo]: ...



from nortl import Const, Engine
from nortl.utils.operand_extraction import extract_operands


def test_extract_operands() -> None:
    e = Engine('my_engine')

    a = e.define_input('A')
    b = e.define_input('B')

    c = e.define_input('C')

    con = Const(25, 8)

    res = [r.render() for r in extract_operands((a + b) * c)]

    assert a.render() in res
    assert b.render() in res
    assert c.render() in res
    assert con.render() not in res

    res = [r.render() for r in extract_operands((a + b) * c * con, keep_const=False)]

    assert a.render() in res
    assert b.render() in res
    assert c.render() in res
    assert con.render() not in res

    res = [r.render() for r in extract_operands((a + b) * c * con, keep_const=True)]

    assert a.render() in res
    assert b.render() in res
    assert c.render() in res
    assert con.render() in res


def test_extract_operands_slice_handling() -> None:
    e = Engine('my_engine')

    a = e.define_input('A', 8)

    res = [r.render() for r in extract_operands(a[3:0])]

    assert res == ['A[3:0]']


def test_extract_operands_scratch() -> None:
    e = Engine('my_engine')

    a = e.define_scratch(8)

    res = [r.render() for r in extract_operands(a[3:0])]

    assert res == ['SCRATCH_SIGNAL[3:0]']



from typing import Dict, Final, List, Optional, Union

from .operations import Const
from .protocols import ModuleProto, ParameterProto, PermanentSignal, Renderable


class Module:
    """Representation of a Verilog module."""

    def __init__(self, name: str, hdl_code: str = '') -> None:
        """Initialize a new Verilog module.

        Arguments:
            name: Name of the module.
            hdl_code: Verilog/VHDL code of the module that should be included in the output.
        """
        self._name: str = name
        self._port_names: List[str] = []
        self._parameters: Dict[str, int] = {}
        self._hdl_code: str = hdl_code

        # The following variable is used by the verilog renderer and masks out the connections of clk and reset
        # Should be handled with care! Therefore this will not be exposed
        self._ignore_clk_rst_connection: bool = False

        # This variable is used for synthesizing the clock network in the final module.
        # The verilog renderer will ignore this signal, if it is set to 'None', otherwise, it will
        # create and connect a signal and wire it up to the enable of the clock gate of the top module.
        self._clk_request_port: Union[None, str] = None

    @property
    def name(self) -> str:
        """Module name."""
        return self._name

    @property
    def ports(self) -> List[str]:
        """List of ports for this module."""
        return self._port_names

    @property
    def parameters(self) -> Dict[str, int]:
        """Dictionary of parameter names and their values."""
        return self._parameters

    @property
    def hdl_code(self) -> str:
        """HDL Code of the module."""
        return self._hdl_code

    def add_port(self, port_name: str) -> None:
        """Add a new port to the module.

        Arguments:
            port_name (str): The signal representing the port.
        """
        self._port_names.append(port_name)

    def has_port(self, port_name: str) -> bool:
        """Test, if a given port (identivfied by name) is in this module.

        Arguments:
            port_name (str): Port name to be checked

        Returns:
            bool: True, if port is in module
        """
        return port_name in self._port_names

    def add_parameter(self, name: str, value: int) -> None:
        """Add a new parameter to the module.

        Arguments:
            name: Name of the parameter.
            value: Value of the parameter.
        """
        self._parameters[name] = value

    def set_clk_request(self, port_name: str) -> None:
        """Set the clock request port for the rendering process.

        Arguments:
           port_name (str): The name of the clock request port.

        Note that this method should be called after all ports have been added.
        """
        if port_name in self._port_names:
            self._clk_request_port = port_name
        else:
            raise ValueError(f'Port {port_name} not found in module ports.')

    @property
    def clk_request_port(self) -> Optional[str]:
        """Get the clock request port for the rendering process.

        Returns:
           Optional[str]: The name of the clock request port or None if not set.
        """
        return self._clk_request_port


class ModuleInstance:
    """Representation of an instance of a Verilog module."""

    is_primitive: Final = False

    def __init__(self, module: ModuleProto, name: str, clock_gating: bool = False) -> None:
        """Initialize a new instance of a Verilog module.

        Arguments:
            module: The module to be instantiated.
            name: Name of the instance.
            clock_gating: If the clock is gated if the current state demands it (and clock gating is enabled in renderer)
        """
        self._module = module
        self._name = name
        self._port_connections: Dict[str, PermanentSignal] = {}  # Use signal name as key
        self._parameter_overrides: Dict[str, Union[Const, ParameterProto, Renderable]] = {}
        self._clock_gating = clock_gating

        if self._module.clk_request_port is not None:
            self._clock_gating = True

    @property
    def module(self) -> ModuleProto:
        """The module being instantiated."""
        return self._module

    @property
    def name(self) -> str:
        """Name of the instance."""
        return self._name

    @property
    def port_connections(self) -> Dict[str, PermanentSignal]:
        """Connections between the ports and signals."""
        return self._port_connections

    @property
    def parameter_overrides(self) -> Dict[str, Union[Const, ParameterProto, Renderable]]:
        """Dictionary of overridden parameter names and their values."""
        return self._parameter_overrides

    def connect_port(self, port_name: str, signal: PermanentSignal) -> None:
        """Connect a port of the instance to a signal.

        Arguments:
            port_name: The port being connected.
            signal: The signal it is being connected to.
        """
        if port_name not in self.module.ports:
            raise ValueError(f'Port {port_name} does not exist in module {self.module.name}')

        self._port_connections[port_name] = signal  # Use signal name as key

    def get_connected_signal(self, port_name: str) -> PermanentSignal:
        """Returns the signal that is connected to the given port.

        Arguments:
            port_name (str): The port where the connected signal should be acquired
        """

        if port_name not in self.module.ports:
            raise ValueError(f'Port {port_name} does not exist in module {self.module.name}')

        return self._port_connections[port_name]

    def override_parameter(self, name: str, value: Union[int, ParameterProto, Renderable]) -> None:
        """Override a parameter of the instance.

        Arguments:
            name: Name of the parameter to be overridden.
            value: New value for the parameter.
        """
        if name not in self.module.parameters:
            raise ValueError(f'Parameter {name} does not exist in module {self.module.name}')

        if isinstance(value, int):
            self._parameter_overrides[name] = Const(value)
        else:
            self._parameter_overrides[name] = value



from abc import ABC, abstractmethod
from math import ceil, log2
from typing import List, Literal, Optional

from nortl.core.protocols import WorkerProto

from .process import AlwaysFF, VerilogAssignment, VerilogCase, VerilogIf
from .structural import VerilogDeclaration
from .utils import VerilogRenderable


class StateRegisterBase(ABC):
    """Abstract base class for state register implementations.

    This class defines the common interface for state register implementations,
    allowing for different state encoding strategies while maintaining a consistent
    API. Subclasses can override specific methods to implement different behaviors.
    """

    def __init__(self, worker: WorkerProto) -> None:
        """Initialize the state register.

        Args:
            worker: The worker containing the states.
        """
        self.worker = worker
        self.state_var = worker.create_scoped_name('state')
        self.next_state_var = worker.create_scoped_name('state_nxt')
        self._current_case: VerilogCase | None = None
        self.states: List[str] = []

    @abstractmethod
    def declare(self) -> VerilogDeclaration:
        """Declares the state register in the Verilog module.

        This method creates the state and next_state variables and adds all state
        members to the enumeration. Subclasses may override this to customize
        the declaration format.

        Returns:
            A VerilogDeclaration object for the state register.
        """
        pass

    @abstractmethod
    def new_case(self, reg: Literal['next state', 'current state']) -> None:
        """Creates a new case statement for the specified state.

        Args:
            reg: Either 'next state' or 'current state' to determine which variable
                 the case statement will use.
        """
        pass

    @abstractmethod
    def add_case(self, case_value: str, item: VerilogRenderable) -> None:
        """Adds an item to the current case.

        Args:
            case_value: The state name for the case statement.
            item: The Verilog item to add (assignment, if statement, print, etc.).

        Raises:
            ValueError: If no active case exists. Call new_case() first or if case_value has not been introduced via add_state
        """
        pass

    @abstractmethod
    def add_state(self, state_name: str) -> None:
        """Adds a state name to the list of states.

        Args:
            state_name: The name of the state to add.
        """
        pass

    @abstractmethod
    def build_case(self) -> VerilogCase:
        """Builds and returns the VerilogCase object from accumulated items.

        Returns:
            A VerilogCase object containing all items added via add_case().

        Raises:
            ValueError: If no active case exists. Call new_case() first.
        """
        pass

    @abstractmethod
    def build_state_transition(self) -> AlwaysFF:
        """Builds the state transition logic.

        This method generates the logic for transitioning the engine from one state
        to another, including reset handling and sync reset conditions. Subclasses
        may override this to customize the transition logic.

        Returns:
            An AlwaysFF object containing the state transition logic.
        """
        pass

    @abstractmethod
    def state_transition(self, next_state: Optional[str]) -> VerilogAssignment:
        """Realzies a transition to a next state."""


class StateRegister(StateRegisterBase):
    """Handles state register declaration and case statement assembly for Verilog generation.

    This class isolates state encoding logic, allowing for easy experimentation with
    different state encodings without modifying the renderer's core logic.
    """

    def __init__(self, worker: WorkerProto) -> None:
        """Initialize the StateRegister.

        Args:
            worker: The worker containing the states.
        """
        super().__init__(worker)

    def declare(self) -> VerilogDeclaration:
        """Declares the state register (enum) in the Verilog module.

        This method creates the state and next_state variables as an enum type
        and adds all state members to the enumeration.
        """
        state_var = self.state_var
        state_nxt_var = self.next_state_var
        state_var_decl = VerilogDeclaration('enum', [state_var, state_nxt_var])
        for state in self.worker.states:
            state_var_decl.add_member(state.name)

        return state_var_decl

    def new_case(self, reg: Literal['next state', 'current state']) -> None:
        """Creates a new case statement for the specified state."""
        if reg == 'next state':
            self._current_case = VerilogCase(self.next_state_var)
        else:
            self._current_case = VerilogCase(self.state_var)

    def add_case(self, case_value: str, item: VerilogRenderable) -> None:
        """Adds an item to the current case.

        Args:
            case_value: The state name for the case statement
            item: The Verilog item to add (assignment, if statement, print, etc.).
        """
        if self._current_case is None:
            raise ValueError('No active case. Call new_case() first.')
        if case_value not in self.states:
            raise ValueError(f'Value {case_value} not found. Please use add_state first to introduce it!')

        # Get the first (and only) case value
        self._current_case.add_item(case_value, item)

    def add_state(self, state_name: str) -> None:
        self.states.append(state_name)

    def build_case(self) -> VerilogCase:
        """Builds and returns the VerilogCase object from accumulated items.

        Returns:
            A VerilogCase object containing all items added via add_case().
        """
        if self._current_case is None:
            raise ValueError('No active case. Call new_case() first.')

        return self._current_case

    def build_state_transition(self) -> AlwaysFF:
        """Builds the state transition logic.

        This method generates the logic for transitioning the engine from one state to another,
        including reset handling and sync reset conditions.
        """
        state_transition = AlwaysFF('CLK_I', 'RST_ASYNC_I')
        state_transition.add_reset(VerilogAssignment(self.state_var, self.worker.reset_state.name))

        block = VerilogIf(f'{self.worker.sync_reset}')
        block.true_branch.add(VerilogAssignment(self.state_var, self.worker.reset_state.name))
        block.false_branch.add(VerilogAssignment(self.state_var, self.next_state_var))

        state_transition.add(block)

        return state_transition

    def encode(self, state_name: str) -> str:
        """Encodes the state in the currently realized way."""
        return state_name

    def state_transition(self, next_state: Optional[str]) -> VerilogAssignment:
        if next_state is None:
            return VerilogAssignment(self.next_state_var, self.state_var)
        return VerilogAssignment(self.next_state_var, next_state)


class OneHotEncodedStateRegister(StateRegisterBase):
    """Handles state register declaration and case statement assembly for Verilog generation     using one-hot encoding.

    This class isolates state encoding logic, allowing for easy experimentation with
    different state encodings without modifying the renderer's core logic. Unlike the
    standard StateRegister which uses Systemverilog enums, this implementation uses
    one-hot encoding where each state is represented by a single bit being set to 1.

    Example:
        >>> from nortl.core.protocols import WorkerProto
        >>> from nortl.renderer.verilog_utils import OneHotEncodedStateRegister
        >>> # Create a one-hot encoded state register
        >>> state_reg = OneHotEncodedStateRegister(worker)
        >>> # The state variable will be declared as logic [N-1:0] state
        >>> # where N is the number of states
    """

    def __init__(self, worker: WorkerProto) -> None:
        """Initialize the OneHotEncodedStateRegister.

        Args:
            worker: The worker containing the states.
        """
        super().__init__(worker)
        self.state_width = len(self.worker.states)

    def declare(self) -> VerilogDeclaration:
        """Declares the state register (one-hot encoded logic) in the Verilog module.

        This method creates the state and next_state variables as logic types with
        a width equal to the number of states. Each state corresponds to a single bit
        in the register (one-hot encoding).

        Returns:
            A VerilogDeclaration object for the state register.
        """
        state_var = self.state_var
        state_nxt_var = self.next_state_var
        state_var_decl = VerilogDeclaration('logic', [state_var, state_nxt_var], self.state_width)

        return state_var_decl

    def new_case(self, reg: Literal['next state', 'current state']) -> None:
        """Creates a new case statement for the specified state.

        Args:
            reg: Either 'next state' or 'current state' to determine which variable
                 the case statement will use.
        """
        if reg == 'next state':
            self._current_case = VerilogCase(self.next_state_var, 'unique')
        else:
            self._current_case = VerilogCase(self.state_var, 'unique')

    def add_case(self, case_value: str, item: VerilogRenderable) -> None:
        """Adds an item to the current case.

        Args:
            case_value: The state name for the case statement.
            item: The Verilog item to add (assignment, if statement, print, etc.).

        Raises:
            ValueError: If no active case exists. Call new_case() first or if case_value has not been introduced via add_state
        """
        if self._current_case is None:
            raise ValueError('No active case. Call new_case() first.')
        if case_value not in self.states:
            raise ValueError(f'Value {case_value} not found. Please use add_state first to introduce it!')

        # Get the index of the state and create the one-hot pattern
        state_encoded = list('0' * len(self.states))
        state_encoded[self.states.index(case_value)] = '1'

        one_hot_pattern = f"{len(self.states)}'b{''.join(state_encoded)}"

        # Get the first (and only) case value
        self._current_case.add_item(one_hot_pattern, item)

    def add_state(self, state_name: str) -> None:
        """Adds a state name to the list of states.

        Args:
            state_name: The name of the state to add.
        """
        self.states.append(state_name)

    def build_case(self) -> VerilogCase:
        """Builds and returns the VerilogCase object from accumulated items.

        Returns:
            A VerilogCase object containing all items added via add_case().

        Raises:
            ValueError: If no active case exists. Call new_case() first.
        """
        if self._current_case is None:
            raise ValueError('No active case. Call new_case() first.')

        # Add missing keys to case object
        for statename in self.states:
            if self.encode(statename) not in self._current_case.cases.keys():
                self._current_case.add_case(self.encode(statename))

        return self._current_case

    def build_state_transition(self) -> AlwaysFF:
        """Builds the state transition logic.

        This method generates the logic for transitioning the engine from one state
        to another, including reset handling and sync reset conditions. The next state
        is computed using a one-hot encoded logic where each state's next state is
        determined by checking which current state bit is set.

        Returns:
            An AlwaysFF object containing the state transition logic.
        """
        state_transition = AlwaysFF('CLK_I', 'RST_ASYNC_I')
        state_transition.add_reset(VerilogAssignment(self.state_var, self.encode(self.worker.reset_state.name)))

        # Build the next state logic using a case statement on the one-hot encoded state

        block = VerilogIf(f'{self.worker.sync_reset}')
        block.true_branch.add(VerilogAssignment(self.state_var, self.encode(self.worker.reset_state.name)))
        block.false_branch.add(VerilogAssignment(self.state_var, self.next_state_var))

        state_transition.add(block)

        return state_transition

    def encode(self, state_name: str) -> str:
        """Encodes the state in the currently realized way."""
        state_encoded = list('0' * len(self.states))
        state_encoded[self.states.index(state_name)] = '1'

        return f"{len(self.states)}'b{''.join(state_encoded)}"

    def state_transition(self, next_state: Optional[str]) -> VerilogAssignment:
        if next_state is None:
            return VerilogAssignment(self.next_state_var, self.state_var)
        return VerilogAssignment(self.next_state_var, self.encode(next_state))


class MultiHotEncodedStateRegister(StateRegisterBase):
    """Handles state register declaration and case statement assembly for Verilog generation using multi-hot encoding.

    This class isolates state encoding logic, allowing for easy experimentation with
    different state encodings without modifying the renderer's core logic. Unlike the
    standard StateRegister which uses Systemverilog enums, this implementation uses
    multiple one-hot encoded registers where each state is represented by multiple
    one-hot encoded packets.

    Each state ID is split into N-bit packets, and each packet is one-hot encoded.
    For example, with packet_size=2 and 6 states:
    - State 0 (00): packet0=01, packet1=01 (one-hot for 0 in each 2-bit packet)
    - State 1 (01): packet0=10, packet1=01 (one-hot for 1 in first packet, 0 in second)
    - State 2 (10): packet0=01, packet1=10 (one-hot for 0 in first packet, 2 in second)
    - etc.

    Example:
        >>> from nortl.core.protocols import WorkerProto
        >>> from nortl.renderer.verilog_utils import MultiHotEncodedStateRegister
        >>> # Create a multi-hot encoded state register with 2-bit packets
        >>> state_reg = MultiHotEncodedStateRegister(worker, packet_size=2)
        >>> # The state variable will be declared as logic [N*packet_size-1:0] state
        >>> # where N is the number of packets needed to represent all states
    """

    def __init__(self, worker: WorkerProto, packet_size: int = 4) -> None:
        """Initialize the MultiHotEncodedStateRegister.

        Args:
            worker: The worker containing the states.
            packet_size: The size of each one-hot encoded packet in bits.
                        Each packet can represent up to 2^packet_size values.
        """
        super().__init__(worker)
        self.packet_size = packet_size
        self.num_states = len(self.worker.states)
        # Calculate number of packets needed: ceil(log2(num_states) / packet_size)
        bits_needed = ceil(log2(self.num_states)) if self.num_states > 1 else 1
        self.num_packets = ceil(bits_needed / packet_size)
        # Total width = num_packets * packet_size
        self.state_width = self.num_packets * (2**packet_size)

    def declare(self) -> VerilogDeclaration:
        """Declares the state register (multi-hot encoded logic) in the Verilog module.

        This method creates the state and next_state variables as logic types with
        a width equal to num_packets * packet_size. Each state is represented by
        multiple one-hot encoded packets.

        Returns:
            A VerilogDeclaration object for the state register.
        """
        state_var = self.state_var
        state_nxt_var = self.next_state_var
        state_var_decl = VerilogDeclaration('logic', [state_var, state_nxt_var], self.state_width)

        return state_var_decl

    def new_case(self, reg: Literal['next state', 'current state']) -> None:
        """Creates a new case statement for the specified state.

        Args:
            reg: Either 'next state' or 'current state' to determine which variable
                 the case statement will use.
        """
        if reg == 'next state':
            self._current_case = VerilogCase(self.next_state_var, 'unique')
        else:
            self._current_case = VerilogCase(self.state_var, 'unique')

    def add_case(self, case_value: str, item: VerilogRenderable) -> None:
        """Adds an item to the current case.

        Args:
            case_value: The state name for the case statement.
            item: The Verilog item to add (assignment, if statement, print, etc.).

        Raises:
            ValueError: If no active case exists. Call new_case() first or if case_value has not been introduced via add_state
        """
        if self._current_case is None:
            raise ValueError('No active case. Call new_case() first.')
        if case_value not in self.states:
            raise ValueError(f'Value {case_value} not found. Please use add_state first to introduce it!')

        # Get the index of the state and create the multi-hot pattern
        multi_hot_pattern = self.encode(case_value)

        # Get the first (and only) case value
        self._current_case.add_item(multi_hot_pattern, item)

    def add_state(self, state_name: str) -> None:
        """Adds a state name to the list of states.

        Args:
            state_name: The name of the state to add.
        """
        self.states.append(state_name)

    def build_case(self) -> VerilogCase:
        """Builds and returns the VerilogCase object from accumulated items.

        Returns:
            A VerilogCase object containing all items added via add_case().

        Raises:
            ValueError: If no active case exists. Call new_case() first.
        """
        if self._current_case is None:
            raise ValueError('No active case. Call new_case() first.')

        # Add missing keys to case object
        for statename in self.states:
            if self.encode(statename) not in self._current_case.cases.keys():
                self._current_case.add_case(self.encode(statename))

        return self._current_case

    def build_state_transition(self) -> AlwaysFF:
        """Builds the state transition logic.

        This method generates the logic for transitioning the engine from one state
        to another, including reset handling and sync reset conditions. The next state
        is computed using a multi-hot encoded logic where each state's next state is
        determined by checking which current state bits are set.

        Returns:
            An AlwaysFF object containing the state transition logic.
        """
        state_transition = AlwaysFF('CLK_I', 'RST_ASYNC_I')
        state_transition.add_reset(VerilogAssignment(self.state_var, self.encode(self.worker.reset_state.name)))

        # Build the next state logic using a case statement on the multi-hot encoded state

        block = VerilogIf(f'{self.worker.sync_reset}')
        block.true_branch.add(VerilogAssignment(self.state_var, self.encode(self.worker.reset_state.name)))
        block.false_branch.add(VerilogAssignment(self.state_var, self.next_state_var))

        state_transition.add(block)

        return state_transition

    def encode(self, state_name: str) -> str:
        """Encodes the state in the currently realized way.

        The state ID is split into N-bit packets, and each packet is one-hot encoded.
        For a packet_size-bit one-hot encoding, value v has bit v set to 1, others to 0.

        Example with packet_size=2:
        - State 0 (ID=0): packet0=01 (one-hot for 0), packet1=01 (one-hot for 0)
        - State 1 (ID=1): packet0=10 (one-hot for 1), packet1=01 (one-hot for 0)
        - State 2 (ID=2): packet0=01 (one-hot for 0), packet1=10 (one-hot for 1)
        """
        state_index = self.states.index(state_name)
        packets = []

        for packet_num in range(self.num_packets):
            # Extract the packet value from the state index
            # Each packet represents packet_size bits of the state ID
            packet_value = (state_index >> (packet_num * self.packet_size)) & ((1 << self.packet_size) - 1)

            case_str = bin(2**packet_value)[2:].zfill(2**self.packet_size)

            packets.append(f"{2**self.packet_size}'b{case_str}")

        # Concatenate packets: most significant packet first
        return '{' + ', '.join(reversed(packets)) + '}'

    def state_transition(self, next_state: Optional[str]) -> VerilogAssignment:
        if next_state is None:
            return VerilogAssignment(self.next_state_var, self.state_var)
        return VerilogAssignment(self.next_state_var, self.encode(next_state))



"""noRTL Code Generation Engine.

This module provides the main Engine class and mixins for creating noRTL-based code generation.
It includes components for communication, constructs for flow control, and algorithms for analysis.
"""

from contextlib import contextmanager
from typing import Iterator, Union

from nortl import verilog_library
from nortl.algorithms import EmptyStateRemovalMixin, ReachabilityAnalysisMixin, StateMergerMixin
from nortl.components import Channel, ElasticChannel, Timer
from nortl.core import All, Any, Concat, Const, CoreEngine, IfThenElse, Var, Volatile, enable_tracing, to_renderable
from nortl.core.constructs import Condition, ElseCondition, Fork, ForLoop, Segment, WhileLoop
from nortl.core.protocols import ParameterProto, Renderable

__all__ = [
    'All',
    'Any',
    'Concat',
    'Const',
    'CoreEngine',
    'Engine',
    'IfThenElse',
    'Segment',
    'Var',
    'Volatile',
    'enable_tracing',
    'to_renderable',
]


class ComponentsMixin(CoreEngine):
    """Mixin class providing noRTL Components.

    This allows easier access to the components, without needing to import them.
    """

    def create_channel(self, width: int, name: str = 'channel') -> Channel:
        """Create a channel.

        Example:
        ```python
        engine = Engine("my_engine")
        channel = engine.create_channel(8, "data_channel")
        ```

        Arguments:
            width: Bit width of the channel.
            name: Name of the channel instance.
        """
        return Channel(self, width, name=name)

    def create_elastic_channel(self, width: int, name: str = 'channel') -> ElasticChannel:
        """Create an elastic channel.

        Example:
        ```python
        engine = Engine("my_engine")
        channel = engine.create_elastic_channel(8, "data_channel")
        ```

        Arguments:
            width: Bit width of the channel.
            name: Name of the channel instance.
        """
        return ElasticChannel(self, width, name=name)

    def create_timer(self, width: Union[int, ParameterProto] = 16, instance_name_prefix: str = 'I_TIMER', clock_gating: bool = False) -> Timer:
        """Create a timer.

        Example:
        ```python
        engine = Engine("my_engine")
        timer = engine.create_timer(32, "I_TIMER", clock_gating=True)
        ```

        Arguments:
            width: Bit width of the timer counter (default: 16).
            instance_name_prefix: Prefix for the instance name (default: 'I_TIMER').
            clock_gating: Enable clock gating for the timer (default: False).
        """
        return Timer(self, width=width, instance_name_prefix=instance_name_prefix, clock_gating=clock_gating)


class ConstructsMixin(CoreEngine):
    """Mixin class providing noRTL Constructs.

    This allows easier access to the flow control constructs, without needing to import them.
    """

    def condition(self, condition: Renderable) -> Condition:
        """Adds a Condition.

        Example:
        ```python
        engine = Engine("my_engine")
        in_signal = engine.define_input("IN")

        with engine.condition(in_signal == 1):
            engine.sync()
            engine.do_something()
        ```

        Arguments:
            condition: A Renderable signal representing the condition to evaluate.
        """
        return Condition(self, condition)

    def else_condition(self) -> ElseCondition:
        """Adds an Else Condition.

        The else condition is automatically derived from the previous conditions in the state.

        Example:
        ```python
        engine = Engine("my_engine")
        in_signal = engine.define_input("IN")

        with engine.condition(in_signal == 1):
            engine.sync()
            engine.do_something()
        with engine.else_condition():
            engine.do_something_else()
        ```

        """
        return ElseCondition(self)

    def fork(self, threadname: str) -> Fork:
        """Adds a Fork.

        Example:
        ```python
        engine = Engine("my_engine")
        with engine.fork("thread1"):
            engine.do_something()
        ```

        Arguments:
            threadname: Name of the fork thread.
        """
        return Fork(self, threadname)

    def for_loop(
        self, start: Union[Renderable, int], stop: Union[Renderable, int], step: Union[Renderable, int] = Const(1), counter_width: int = 16
    ) -> ForLoop:
        """Adds a For loop.

        Example:
        ```python
        engine = Engine("my_engine")
        out = engine.define_output("test_output", width=8)

        with engine.for_loop(0, 100, 2) as i:
            engine.set(out, i)
        ```

        Arguments:
            start: A signal or int representing the start value of the loop counter.
            stop: A signal or int representing the final value of the loop counter (non-inclusive).
            step: A signal or int representing the step value of the counter (default: 1).
            counter_width: Width of the counter variable (default: 16).
        """
        return ForLoop(self, start, stop, step=step, counter_width=counter_width)

    def while_loop(self, condition: Renderable) -> WhileLoop:
        """Adds a While loop.

        Example:
        ```python
        engine = Engine("my_engine")
        out = engine.define_output("test_output", width=8)

        with engine.while_loop(out < 4) as _:
            engine.set(out, out + 1)
        ```

        Arguments:
            condition: A signal or expression representing the condition of the while loop.
        """
        return WhileLoop(self, condition)


class ManagementMixin(CoreEngine):
    """Mixin Class that provides context managers with convenience functions."""

    @contextmanager
    def context(self) -> Iterator[None]:
        """Creates an empty context for scoping scratch variables."""
        self.scratch_manager.enter_context()
        yield
        self.scratch_manager.exit_context()

    @contextmanager
    def collapse_sync(self) -> Iterator[None]:
        """Marks all states as collapsable. This means, that all empty states (engine.sync() without assignments) may be removed by an optimzation."""
        # Use a counter to handle nested contexts
        if 'collapse_sync_depth' not in self.state_metadata_template:
            self.state_metadata_template['collapse_sync_depth'] = 0
        self.state_metadata_template['collapse_sync_depth'] += 1
        if self.state_metadata_template['collapse_sync_depth'] == 1:
            self.state_metadata_template['collapsable'] = True
        yield
        self.state_metadata_template['collapse_sync_depth'] -= 1
        if self.state_metadata_template['collapse_sync_depth'] == 0:
            self.state_metadata_template['collapsable'] = False


class Engine(
    ComponentsMixin,
    ConstructsMixin,
    ManagementMixin,
    ReachabilityAnalysisMixin,
    StateMergerMixin,
    EmptyStateRemovalMixin,
    CoreEngine,
):
    """noRTL Engine.

    The Engine class is the main entry point for creating noRTL-based code generation.
    It provides access to components, constructs, and algorithms through mixin classes.

    Example:
    ```python
    from nortl import Engine

    engine = Engine("my_engine")
    in_signal = engine.define_input("IN")
    out_signal = engine.define_output("OUT", width=8)

    with engine.condition(in_signal == 1):
        engine.set(out_signal, 5)
    ```
    """

    def __init__(self, module_name: str, reset_state_name: str = 'IDLE') -> None:
        """Initialize a new noRTL Engine.

        Arguments:
            module_name: Name of the resulting SystemVerilog module.
            reset_state_name: Default name for the reset state (default: 'IDLE').
        """
        super().__init__(module_name, reset_state_name)

        for module in verilog_library.get_modules():
            self.add_module(module)



from nortl import Const, Engine


def test_unreachable_transitions_not_added() -> None:
    e = Engine('my_engine')

    e.sync()

    c = Const(1)

    e.jump_if(c == 0, e.reset_state)  # Transition that is never taken

    assert len(e.current_state.transitions) == 0


def test_state_metadata_template() -> None:
    """Test that states created via engine.sync() inherit metadata from state_metadata_template."""
    e = Engine('test_engine')

    # Set the state metadata template
    e.state_metadata_template = {'key1': 'value1', 'key2': 42}

    # Create states using sync()
    e.sync()

    # Verify that each state has the metadata from the template
    state = e.current_state
    assert state.has_metadata('key1')
    assert state.get_metadata('key1') == 'value1'
    assert state.has_metadata('key2')
    assert state.get_metadata('key2') == 42



from nortl import Engine


def test_state_merger_no_merge() -> None:
    e = Engine('my_engine')

    for _ in range(10):
        e.sync()

    e.state_merging()


def test_state_merger() -> None:
    # State machine with merges
    e = Engine('my_engine')
    a = e.define_input('A')
    b = e.define_input('B')

    with e.condition(a):
        for _ in range(10):
            e.sync()

    with e.condition(b):
        for _ in range(10):
            e.sync()

    e.state_merging()

    # State machine without merges
    e2 = Engine('my_engine')
    a = e2.define_input('A')
    b = e2.define_input('B')

    with e2.condition(a):
        for _ in range(10):
            e2.sync()

    assert len(e.main_worker.states) == len(e2.main_worker.states)



from typing import Dict, Iterator, List, Optional, Tuple

from nortl.core.protocols import EngineProto


class MermaidRenderer:
    """This class contains the methods to render a nortl to a mermaid state diagram.

    Note that only states and transitions are shown. For transitions, the conditions are omitted.
    It's purpose is to support illustration only!
    """

    # FIXME: Include Fork/Joins

    def __init__(self, engine: EngineProto):
        self.engine = engine
        self.indent_level = 0

    def render(self) -> str:
        ret = '---\n'
        ret += f'title {self.engine.module_name}\n'
        ret += '---\n'
        ret += 'stateDiagram-v2\n'

        for _, worker_graph in self._render_workers():
            ret += worker_graph
        return ret

    def render_workers(self) -> Iterator[Tuple[str, str]]:
        for worker, worker_graph in self._render_workers():
            ret = '---\n'
            ret += f'title {self.engine.module_name} {worker}\n'
            ret += '---\n'
            ret += 'stateDiagram-v2\n'
            ret += worker_graph
            yield worker, ret

    def _render_workers(self) -> Iterator[Tuple[str, str]]:  # noqa: C901
        for worker, worker_states in self.engine.states.items():
            transitions: List[Tuple[str, str, str]] = []
            assignments: Dict[str, List[Tuple[str, str, Optional[str]]]] = {}

            for state in worker_states:
                for condition, next_state in state.transitions:
                    transitions.append((state.name, next_state.name, condition.render()))

                item: List[Tuple[str, str, Optional[str]]] = []
                for assignment in state.assignments:
                    if assignment.unconditional:
                        item.append((assignment.signal.render(), assignment.value.render(), None))
                    else:
                        for condition, value in assignment.flatten_cases():
                            item.append((assignment.signal.render(), value.render(), condition.render()))
                assignments[state.name] = item

            # Generate the mermaid code
            ret = ''
            for s1, s2, cond in transitions:
                ret += f'    {s1} --> {s2}: {cond.replace(":", "_")}\n'

            for s1, assignmentlist in assignments.items():
                if len(assignmentlist) != 0:
                    ret += f'    note right of {s1}\n'
                    for reg, val, cond_ in assignmentlist:
                        if cond_ is None:
                            ret += f'        set {reg.replace(":", "_")} to {val.replace(":", "_")}\n'
                        else:
                            ret += f'        if {cond_.replace(":", "_")}, set {reg.replace(":", "_")} to {val.replace(":", "_")}\n'

                    ret += '    end note\n'

            yield worker, ret



"""Miscellaneous operations."""

from typing import Optional, TypeVar, Union

from nortl.core.protocols import Operand, Renderable
from nortl.core.renderers.operations.base import SingleRenderer, const_unpack

__all__ = [
    'Inversion',
    'Negative',
    'Positive',
]


class Negative(SingleRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        return f'-{self.value}'

    @staticmethod
    def eval(value: Operand) -> Optional[int]:
        """Evaluate operation into constant value."""
        if (value_ := const_unpack(value)) is not None:
            return -value_
        return None


T = TypeVar('T', bound=Renderable)


class Positive(SingleRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        # Positive renderer should never really be used under normal conditions, because the operation does nothing
        return f'+{self.value}'  # pragma: no cover

    @staticmethod
    def eval(value: T) -> Optional[Union[T, int]]:
        """Evaluate operation into constant value or single operand."""
        if (value_ := const_unpack(value)) is not None:
            return value_

        # Always return the input operand, no need to wrap it in a Operation
        return value


class Inversion(SingleRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        return f'~({self.value})'

    @staticmethod
    def eval(value: Operand) -> Optional[int]:
        """Evaluate operation into constant value."""
        if (value_ := const_unpack(value)) is not None:
            return ~value_
        return None



import re
import subprocess
from abc import ABC, abstractmethod
from inspect import currentframe, getframeinfo
from pathlib import Path
from tempfile import TemporaryDirectory
from types import FrameType

import pytest

from nortl.core import Const
from nortl.core.constructs import Condition, ElseCondition, Fork
from nortl.core.modifiers import UnregisteredRead
from nortl.core.operations import to_renderable
from nortl.core.protocols import EngineProto, Renderable
from nortl.renderer.verilog_renderer import VerilogRenderer


class NoRTLTestBase[T: EngineProto](ABC):
    """Base class for noRTL tests with Verilog simulation.

    Subclasses must implement:
        - build_engine(): Create and configure the engine to be tested
        - generate_testbench(): Generate the Verilog testbench file

    The simulation is automatically run before each test method.
    Test methods can access:
        - self.engine: The engine instance
        - self.simulation_result: The subprocess result from simulation
        - self.simulation_output: Parsed simulation output (string)
    """

    @abstractmethod
    def init_sequence(self) -> T:
        """Initializes the engine that is to be tested.

        Its intention is to create and configure an instance of the Engine that will be tested.

        Example:
        ```python
        class MyTest(NoRTLTestBase):
            def init_sequence(self) -> T:
                engine = Engine("my_test_engine")
                output = engine.define_output("result", width=8)
                return engine
        ```

        Returns:
            Engine: Initialized engine
        """
        pass

    def verify_final_state(self, engine: T) -> None:
        """Verifies the final state of the engine after simulation.

        This state-sequence is called after the dut and testbench threads have ended.
        It can be used to verify the final state of the dut.

        Arguments:
            engine (EngineProto): The engine instance to verify
        """
        pass

    def the_testbench(self, engine: T) -> None:
        """The state sequence that is run in parallel to the engine-under-Test.

        This may use the self.assert* functions to verify the behavior.

        Arguments:
            engine (T): engine instance
        """
        pass

    @abstractmethod
    def dut(self, engine: T) -> None:
        """The dut function is the state sequence that is to be tested.

        It is executed in parallel to the testbench. A call to self.assert* is not forbidden
        but not advised to create a good separation between dut and test code.

        Arguments:
            engine (T): The engine instance
        """
        pass

    def callback_before_rendering(self, engine: T) -> T:
        return engine

    def get_test_engine(self) -> T:
        self.engine = self.init_sequence()

        self.finish_flag = self.engine.define_output('finish', 1, 0)
        self.state_flag = self.engine.define_output('passed', 1, 0)
        self.error_ctr = self.engine.define_local('error_ctr', 16, 0)
        self.timeout = self.engine.define_output('timeout', 1, 0)
        ctr = self.engine.define_local('timeout_ctr', 16, 0)

        self.engine.sync()

        with Fork(self.engine, 'DUT') as t_dut:
            self.dut(self.engine)
        with Fork(self.engine, 'testbench') as t_tb:
            self.the_testbench(self.engine)
        with Fork(self.engine, 'timeout') as t_timeout:
            self.engine.sync()
            self.engine.set(ctr, ctr + 1)
            with Condition(self.engine, ctr == 0xFFFF):
                self.engine.set(self.timeout, 1)
                t_timeout.finish()

        self.engine.sync()

        self.engine.wait_for((t_dut.finished & t_tb.finished) | t_timeout.finished)

        with Condition(self.engine, t_timeout.finished):
            t_dut.cancel()
            t_tb.cancel()
            if not hasattr(self, 'expect_timeout'):
                self.assertTrue(0)  # Show that we had a timeout
            else:
                self.assertTrue(self.expect_timeout)
        with ElseCondition(self.engine):
            t_timeout.cancel()

        self.verify_final_state(self.engine)

        self.engine.sync()

        self.finish_simulation()

        return self.engine

    def increment_error_ctr(self) -> None:
        self.engine.sync()

        with Condition(self.engine, self.error_ctr != 0xFFFF):
            self.engine.set(self.error_ctr, self.error_ctr + 1)
        with ElseCondition(self.engine):
            pass

        self.engine.sync()

    def finish_simulation(self) -> None:
        """Finalizes the simulation by setting flags.

        This method sets the state and finish flags based on the error counter.
        If no errors occurred, the state flag is set to 1 (passed), otherwise 0 (failed).
        """
        self.engine.sync()

        with Condition(self.engine, self.error_ctr == 0):
            self.engine.set(self.state_flag, 1)
        with ElseCondition(self.engine):
            self.engine.set(self.state_flag, 0)

        self.engine.set(self.finish_flag, 1)
        self.engine.sync()

    def print_line(self, frame: FrameType) -> None:
        """Prints an assertion failure message with context.

        Extracts the source code line where the assertion failed and prints it
        along with the file and line number.

        Arguments:
            frame (FrameType): The frame object containing the code context
        """
        fi = getframeinfo(frame)
        if fi.code_context is None:
            return
        code_context = fi.code_context[0]
        code_context = re.sub('#.*', '', code_context)
        code_context = code_context.lstrip().rstrip()
        self.engine.print(f'Assertion \\"{code_context}\\" failed at {fi.filename}:{frame.f_lineno}')

    def assertTrue(self, condition: Renderable | int) -> None:  # noqa: N802
        """Asserts that a condition is true.

        If the condition is false, prints the assertion line and fails the simulation.

        Example:
        ```python
        self.assertTrue(output_signal == 5)
        ```

        Arguments:
            condition (Renderable | int): The condition to assert as true
        """
        if isinstance(condition, int):
            condition = Const(condition, 1)

        with Condition(self.engine, UnregisteredRead(~to_renderable(condition))):
            frame = currentframe()
            if frame is not None:
                frame = frame.f_back
                if frame is not None:
                    self.print_line(frame)
            self.increment_error_ctr()
            self.finish_simulation()
        with ElseCondition(self.engine):
            pass
        self.engine.sync()

    def assertEqual(self, value1: Renderable | int, value2: Renderable | int) -> None:  # noqa: N802
        """Asserts that two values are equal.

        If the values are not equal, prints both values from within the simulation and fails the simulation.

        Example:
        ```python
        self.assertEqual(output_signal, expected_value)
        ```

        Arguments:
            value1 (Renderable | int): The first value to compare
            value2 (Renderable | int): The second value to compare
        """
        with Condition(self.engine, UnregisteredRead(to_renderable(value1) != to_renderable(value2))):
            frame = currentframe()
            if frame is not None:
                frame = frame.f_back
                if frame is not None:
                    self.print_line(frame)

            self.engine.print('Left-hand side: 0x%h', to_renderable(value1))
            self.engine.print('Right-hand side: 0x%h', to_renderable(value2))
            self.increment_error_ctr()
            self.finish_simulation()
        with ElseCondition(self.engine):
            pass
        self.engine.sync()

    def test_compile_and_run(self) -> None:
        """Automatically executed test method.

        This method runs before each test and performs the following steps:

        1. Builds the engine by calling init_sequence()
        2. Generates and compiles Verilog files
        3. Runs the simulation
        4. Makes results available to test methods

        The method raises pytest.fail if the simulation fails.
        """
        # Create engine from derived class
        self.engine = self.callback_before_rendering(self.get_test_engine())

        with TemporaryDirectory() as tempdir_str:
            tmp_path = Path(tempdir_str)

            # Generate Verilog files
            self.verilog_file: Path = tmp_path / 'engine.sv'
            self.testbench_file: Path = Path(__file__).parent / 'templates' / 'testbench.sv'

            renderer = VerilogRenderer(self.engine)
            with open(self.verilog_file, 'w') as fptr:
                fptr.write(renderer.render())

            # Compile with Icarus Verilog
            self.compiled_file: Path = tmp_path / 'engine.vvp'
            compile_result = subprocess.run(
                ['iverilog', '-g2005-sv', '-o', str(self.compiled_file), str(self.verilog_file), str(self.testbench_file)],
                capture_output=True,
                text=True,
            )

            if compile_result.returncode != 0:
                pytest.fail(f'Verilog compilation failed:\n{compile_result.stderr}')

            # Run simulation automatically
            self._run_simulation()
            self.simulation_output: str = self.simulation_result.stdout

            res = re.findall(r'passed=\s*\d+', self.simulation_output)[0]
            print(self.simulation_output)

            pass_state = int(res.split('=')[1]) == 1
            print(pass_state)

            if not pass_state:
                pytest.fail('Verilog simulation discovered errors')

    def _run_simulation(self) -> None:
        """Executes the compiled Verilog simulation.

        This internal method runs the simulation using the vvp simulator.

        Arguments:
            None
        """
        cmd = ['vvp', str(self.compiled_file)]

        self.simulation_result = subprocess.run(cmd, capture_output=True, text=True)



"""This folder holds objects that are instantiated by the CoreEngine to 'manage' certain other objects such as signals."""

from .scratch_manager import ScratchManager
from .signal_manager import SignalManager

__all__ = ['ScratchManager', 'SignalManager']



from nortl import Engine


def test_scratch_release_after_condition_context() -> None:
    """Test that a scratch signal created in a Condition context is released after exiting."""
    engine = Engine('my_engine')
    out = engine.define_output('out', 4, 0)

    with engine.condition(out == 0):
        test = engine.define_scratch(4)
        engine.set(test, 1)

    # The scratch signal should be released after exiting the context
    assert test.released, 'Scratch signal should be released after exiting Condition context'


def test_scratch_release_after_scratch_context() -> None:
    """Test that a scratch signal created in a context is released after exiting."""
    engine = Engine('my_engine')
    engine.sync()

    with engine.context():
        test = engine.define_scratch(4)
        engine.set(test, 1)

    # The scratch signal should be released after exiting the context
    assert test.released, 'Scratch signal should be released after exiting context'


def test_scratch_release_after_else_condition_context() -> None:
    """Test that a scratch signal created in an ElseCondition context is released after exiting."""
    engine = Engine('my_engine')
    out = engine.define_output('out', 4, 0)

    with engine.condition(out == 0):
        engine.set(out, 1)
    with engine.else_condition():
        test = engine.define_scratch(4)
        engine.set(test, 1)

    # The scratch signal should be released after exiting the ElseCondition context
    assert test.released, 'Scratch signal should be released after exiting ElseCondition context'


def test_scratch_release_nested_contexts() -> None:
    """Test that scratch signals in nested contexts are properly released."""
    engine = Engine('my_engine')
    out = engine.define_output('out', 4, 0)

    with engine.condition(out == 0):
        with engine.condition(out == 1):
            test = engine.define_scratch(4)
            engine.set(test, 1)

    # The scratch signal should be released after exiting both nested contexts
    assert test.released, 'Scratch signal should be released after exiting nested Condition contexts'


def test_scratch_manual_release_in_context() -> None:
    """Test that manual release works correctly within the same context."""
    engine = Engine('my_engine')
    out = engine.define_output('out', 4, 0)

    with engine.condition(out == 0):
        test = engine.define_scratch(4)
        engine.set(test, 1)
        test.release()

    # The scratch signal should be manually released
    assert test.released, 'Scratch signal should be manually released'


def test_scratch_not_released_outside_context() -> None:
    """Test that a scratch signal created outside a context is not released when exiting the context."""
    engine = Engine('my_engine')
    out = engine.define_output('out', 4, 0)
    engine.sync()

    test = engine.define_scratch(4)
    engine.set(test, 1)

    with engine.condition(out == 0):
        pass

    # The scratch signal should NOT be released because it was created outside the context
    assert not test.released, "Scratch signal should not be released when exiting a context it wasn't created in"


def test_scratch_context_counter_value() -> None:
    """Test that the context counter is correctly incremented and decremented."""
    engine = Engine('my_engine')
    out = engine.define_output('out', 4, 0)
    engine.sync()

    # Create scratch signal outside context - counter should be 0
    test = engine.define_scratch(4)
    assert test._context_ctr == 0, 'Scratch signal created outside context should have counter 0'

    with engine.condition(out == 0):
        # Inside context - counter should be 1
        assert test._context_ctr == 1, 'Scratch signal inside context should have counter 1'

    # After exiting - counter should be 0 (but not released because it was created outside)
    assert test._context_ctr == 0, 'Scratch signal counter should return to 0 after exiting context'


def test_scratch_release_in_forked_thread() -> None:
    """Test that scratch signals work correctly in forked threads."""
    engine = Engine('my_engine')
    engine.sync()

    # In forked thread, the scratch signal should not be released
    with engine.fork('f1') as f1:
        test = engine.define_scratch(4)
        engine.set(test, 1)
        assert not test.released, 'Scratch signal should not be released in forked thread'

    f1.cancel()

    # After thread ends, the scratch signal should be released
    assert test.released, 'Scratch signal should be released after forked thread ends'



from .mermaid_renderer import MermaidRenderer
from .networkx_renderer import NetworkXRenderer
from .scratchpad_visualization_renderer import ScratchpadVisualizationRenderer
from .verilog_renderer import VerilogRenderer

__all__ = ['MermaidRenderer', 'NetworkXRenderer', 'ScratchpadVisualizationRenderer', 'VerilogRenderer']



"""Signal definition."""

from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from inspect import FrameInfo
from types import TracebackType
from typing import Dict, Final, Generic, Literal, Mapping, Optional, Sequence, Set, Tuple, Type, TypeVar, Union

from typing_extensions import Self

from nortl.core.checker import StaticAccessChecker
from nortl.core.common import NamedEntity, StaticAccess
from nortl.core.exceptions import AccessAfterReleaseError, WriteViolationError
from nortl.core.modifiers import BaseModifier
from nortl.core.operations import OperationTrait
from nortl.core.protocols import (
    ACCESS_CHECKS,
    BIT_ORDER,
    EVENT_TYPES,
    SIGNAL_TYPES,
    AssignmentTarget,
    EngineProto,
    MemoryZoneProto,
    ModuleInstanceProto,
    ParameterProto,
    Renderable,
    SignalProto,
    SignalSliceProto,
    StaticAccessCheckerProto,
    StaticAccessProto,
    ThreadProto,
)
from nortl.utils.type_aliases import IntSlice

__all__ = [
    'ScratchSignal',
    'Signal',
    'SignalSlice',
]

T_Signal = TypeVar('T_Signal', SignalProto, SignalSliceProto)


class ParameterizedEvent:
    """Wrapper object for a parametrized event."""

    def __init__(self, event: EVENT_TYPES):
        self._parameter_dict: OrderedDict[str, Union[int, str, ParameterProto]] = OrderedDict()
        self._event = event

    def __repr__(self) -> str:
        ret = str(self._event)
        for k, v in self._parameter_dict.items():
            ret += f'{k} {v}'

        return ret

    def __hash__(self) -> int:
        return hash(repr(self))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ParameterizedEvent):
            raise NotImplementedError('Can not compare parameterized Events to other types!')
        return repr(self) == repr(other)

    def add_parameter(self, param: str, value: Union[int, str, ParameterProto]) -> None:
        self._parameter_dict[param] = value

    def get_parameter(self, param: str) -> Union[int, str, ParameterProto]:
        return self._parameter_dict[param]


def pick_indexes(all_indexes: Sequence[int], index: Union[int, IntSlice]) -> Sequence[int]:
    """Pick indexes covered by this slice."""
    if isinstance(index, slice):
        # noRTL supports reversed indexes, to better match Verilog. They must be flipped here.
        # In addition, the stop index is treated as inclusive
        start = min(index.start, index.stop)  # type: ignore[type-var]
        stop = max(index.start, index.stop) + 1  # type: ignore[type-var, operator]
        return all_indexes[start:stop]
    else:
        return (all_indexes[index],)


def list_indexes(index: Union[int, IntSlice]) -> Sequence[int]:
    """List all indexes covered by a slice."""
    if isinstance(index, slice):
        # noRTL supports reversed indexes, to better match Verilog. They must be flipped here.
        # In addition, the stop index is treated as inclusive
        start: int = min(index.start, index.stop)  # type: ignore[assignment, type-var]
        stop: int = max(index.start, index.stop)  # type: ignore[assignment, type-var]
        return tuple(range(start, stop + 1))
    else:
        return (index,)


def validate_slice(index: IntSlice) -> Tuple[int, int, BIT_ORDER]:
    """Validate a slice and return start and stop values sorted by size."""
    # The stop value for the Python slice is treated as inclusive
    start, stop, step = index.start, index.stop, index.step

    if start is None:
        raise ValueError('Missing start position for signal slice operation!')
    if stop is None:
        raise ValueError('Missing stop position for signal slice operation!')
    if step is not None:
        raise ValueError('Providing a step size is not supported for signal slice operation!')

    return min(start, stop), max(start, stop), 'L:H' if stop > start else 'H:L'


class _BaseSignal(OperationTrait, NamedEntity, metaclass=ABCMeta):
    """Abstract base class for signals."""

    is_primitive: Final = True
    is_constant: Final = False

    def __init__(self, name: str):
        super().__init__(name)

    @property
    def name(self) -> str:
        """Signal name."""
        return self._name

    # Abstract properties required by base methods
    @property
    @abstractmethod
    def engine(self) -> EngineProto:
        """NoRTL engine that this signal belongs to."""

    @property
    @abstractmethod
    def width(self) -> Union[int, ParameterProto, Renderable]:
        """Signal width in bits."""

    @property
    @abstractmethod
    def escaped_name(self) -> str:
        """Name of signal with any special characters escaped."""

    # Access control
    @property
    @abstractmethod
    def read_accesses(self) -> Set[StaticAccessProto]:
        """Mutable set of read accesses to this signal."""

    @property
    @abstractmethod
    def write_accesses(self) -> Set[StaticAccessProto]:
        """Mutable set of write accesses to this signal."""

    @property
    @abstractmethod
    def last_read_access_thread(self) -> Optional[ThreadProto]:
        """Thread performing the last read access."""

    @last_read_access_thread.setter
    @abstractmethod
    def last_read_access_thread(self, value: ThreadProto) -> None:
        """Thread performing the last read access."""

    @property
    @abstractmethod
    def last_write_access_thread(self) -> Optional[ThreadProto]:
        """Thread performing the last write access."""

    @last_write_access_thread.setter
    @abstractmethod
    def last_write_access_thread(self, value: ThreadProto) -> None:
        """Thread performing the last write access."""

    @property
    @abstractmethod
    def access_checker(self) -> StaticAccessCheckerProto:
        """Static access checker."""

    def write_access(self, ignore: Set[ACCESS_CHECKS] = set()) -> None:
        """Register write access from the current thread.

        If the current thread differs from the last write acess, will invoke the access checker.

        Raises:
            ExclusiveWriteError: If the signals was written by more than one thread.
            NonIdenticalRWError: If the signals was written by one, and read from another thread.
        """
        self.write_accesses.add(StaticAccess(self.engine.current_thread))

        if self.engine.current_thread is not self.last_write_access_thread:
            # Slow check is only executed, if the thread has changed.
            self.access_checker.check(ignore=ignore)

        self.last_write_access_thread = self.engine.current_thread

    def read_access(self, ignore: Set[ACCESS_CHECKS] = set()) -> None:
        """Register read access from the current thread.

        If the current thread differs from the last write acess, will invoke the access checker.

        Raises:
            ExclusiveReadError: If the signal was read from more than one thread.
            NonIdenticalRWError: If the signals was written by one, and read from another thread.
        """
        self.read_accesses.add(StaticAccess(self.engine.current_thread))

        if self.engine.current_thread is not self.last_read_access_thread:
            # Slow check is only executed, if the thread has changed.
            self.access_checker.check(ignore=ignore)

        self.last_read_access_thread = self.engine.current_thread

    def free_access_from_thread(self, thread: ThreadProto) -> None:
        """This function disables all access checks that have their origin in the given thread.

        It is to be used during fork for passing signals to a forked thread. In this case, the signal
        is accessed (written) by the origin thread and handed off to the spawned thread.

        Since parallel running behavior is described below the actual fork context, a colliding access will
        happen after the fork-block has been executed. The collision will be shown once the origin thread
        will access the signal while the forked thread has not ended.
        """
        for access in self.write_accesses | self.read_accesses:
            if access.thread == thread:
                access.disable()


class _AccessControlledSignal(_BaseSignal):
    """Intermediary class for signals that keep track their own access control.

    This is used for signals and scratch signals.
    """

    def __init__(self, name: str) -> None:
        super().__init__(name)

        self._write_accesses: Set[StaticAccessProto] = set()
        self._read_accesses: Set[StaticAccessProto] = set()
        self._last_read_access_thread: Optional[ThreadProto] = None
        self._last_write_access_thread: Optional[ThreadProto] = None
        self._access_checker = StaticAccessChecker(self)

    @property
    def read_accesses(self) -> Set[StaticAccessProto]:
        """Mutable set of read accesses to this signal."""
        return self._read_accesses

    @property
    def write_accesses(self) -> Set[StaticAccessProto]:
        """Mutable set of write accesses to this signal."""
        return self._write_accesses

    @property
    def last_read_access_thread(self) -> Optional[ThreadProto]:
        """Thread performing the last read access."""
        return self._last_read_access_thread

    @last_read_access_thread.setter
    def last_read_access_thread(self, value: ThreadProto) -> None:
        """Thread performing the last read access."""
        self._last_read_access_thread = value

    @property
    def last_write_access_thread(self) -> Optional[ThreadProto]:
        """Thread performing the last write access."""
        return self._last_write_access_thread

    @last_write_access_thread.setter
    def last_write_access_thread(self, value: ThreadProto) -> None:
        """Thread performing the last write access."""
        self._last_write_access_thread = value

    @property
    def access_checker(self) -> StaticAccessCheckerProto:
        """Static access checker."""
        return self._access_checker


class _EventSourceSignal(Generic[T_Signal], _BaseSignal):
    """Class for signals that can be used as the source for a event.

    This is only the case for signals and slice signals, but not for scratch signals, due to their ephemeral nature.
    """

    def __init__(self, name: str) -> None:
        super().__init__(name)

        self._events: Dict[ParameterizedEvent, ModuleInstanceProto] = {}

    @property
    def events(self) -> Mapping[ParameterizedEvent, ModuleInstanceProto]:
        """Events for this signal."""
        return self._events

    def rising(self) -> T_Signal:
        """Create rising edge event."""
        return self._create_edge_detector().get_connected_signal('RISING')  # type: ignore[return-value]

    def falling(self) -> T_Signal:
        """Create falling edge event."""
        return self._create_edge_detector().get_connected_signal('FALLING')  # type: ignore[return-value]

    def delayed(self, cycles: Union[int, ParameterProto] = 1) -> T_Signal:
        """Create event for delayed signal."""
        return self._create_delay(cycles).get_connected_signal('OUT')  # type: ignore[return-value]

    def synchronized(self) -> T_Signal:
        """Create event for synchronized signal."""
        return self._create_synchronized().get_connected_signal('OUT')  # type: ignore[return-value]

    def _create_edge_detector(self) -> ModuleInstanceProto:
        """Creates the edge detector for the signal if it does not exist.

        If the signal has more than one bit, this results in ValueError.
        """
        if self.width != 1:
            raise ValueError('Edge dectors can only be used in 1-bit signals!')

        if (event := ParameterizedEvent('edge')) not in self.events:
            instance_name = f'I_EVENT_EDGE_DETECTOR_{self.escaped_name}'
            instance = self.engine.create_module_instance(module_name='nortl_edge_detector', instance_name=instance_name)
            self._events[ParameterizedEvent('edge')] = instance

            signal_rising = self.engine.define_local(f'EVENT_{self.escaped_name}_rising')
            signal_falling = self.engine.define_local(f'EVENT_{self.escaped_name}_falling')

            self.engine.connect_module_port(instance_name, 'SIGNAL', self)  # type: ignore[arg-type]
            self.engine.connect_module_port(instance_name, 'RISING', signal_rising)
            self.engine.connect_module_port(instance_name, 'FALLING', signal_falling)

            return instance
        else:
            return self.events[event]

    def _create_delay(self, cycles: Union[int, ParameterProto] = 1) -> ModuleInstanceProto:
        """Create a delay for the signal if it does not exist."""
        event = ParameterizedEvent('delay')
        event.add_parameter('cycles', cycles)

        if event not in self.events:
            instance_name = f'I_DELAY_BY_{cycles}_{self.escaped_name}'
            instance = self.engine.create_module_instance(module_name='nortl_delay', instance_name=instance_name)
            self.engine.override_module_parameter(instance_name, 'DELAY_STEPS', cycles)
            self.engine.override_module_parameter(instance_name, 'DATA_WIDTH', self.width)

            self._events[event] = instance

            delayed_signal = self.engine.define_local(f'EVENT_{self.escaped_name}_DELAY_BY_{cycles}', self.width)

            self.engine.connect_module_port(instance_name, 'IN', self)  # type: ignore[arg-type]
            self.engine.connect_module_port(instance_name, 'OUT', delayed_signal)

            return instance
        else:
            return self.events[event]

    def _create_synchronized(self) -> ModuleInstanceProto:
        """Create sync module for this signal if it does not exist."""
        if (event := ParameterizedEvent('sync')) not in self.events:
            instance_name = f'I_SYNC_{self.escaped_name}'
            instance = self.engine.create_module_instance(module_name='nortl_sync', instance_name=instance_name)
            self.engine.override_module_parameter(instance_name, 'DATA_WIDTH', self.width)

            self._events[event] = instance

            delayed_signal = self.engine.define_local(f'EVENT_{self.escaped_name}_SYNCED')

            self.engine.connect_module_port(instance_name, 'IN', self)  # type: ignore[arg-type]
            self.engine.connect_module_port(instance_name, 'OUT', delayed_signal)

            return instance
        else:
            return self.events[event]


class Signal(_AccessControlledSignal, _EventSourceSignal[SignalProto]):
    """Signal definition, representing a Verilog signal.

    Attributes:
        engine: Finite state machine associated with this signal.
        type: The role of the signal (input, output, interface, internal, local).
        name: Name of the signal.
        width: Width in bits of the signal.
        data_type: Data type of the signal. Defaults to 'logic'.

    The signal types are defined as follows:
    * input, output: Port of the module
    * interface: Use a system verilog interface
    * local: A local register used that is not passed to the outside
    * internal: a signal created by internal data structures, not necessarily visible for the user
    """

    def __init__(
        self,
        engine: EngineProto,
        type: SIGNAL_TYPES,
        name: str,
        width: Union[int, ParameterProto, Renderable] = 1,
        data_type: str = 'logic',
        is_synchronized: bool = False,
        pulsing: bool = False,
        assignment: Optional[Renderable] = None,
    ) -> None:
        """Initialize a signal.

        Arguments:
            engine: State machine container object.
            name: Signal name.
            type: Signal type.
            width: Width in bits (default=1).
            data_type: Data type of the signal (default='logic').
            is_synchronized: Indicates, if the signal is synchronous to the local clock domain.
            pulsing: Wether the signal resets automatically to 0 if not written in current state
            assignment: Source expression for combinational assignment.
        """
        if type != 'internal' and name.startswith('_'):
            raise ValueError('Signal names must not start with an underscore!')
        if assignment is not None:
            if type == 'input':
                raise ValueError('Input signals must not have a combinational assignment.')
            if pulsing:
                raise ValueError('Signals with a combinational assignment cannot be pulsing.')

        super().__init__(name)

        self._engine = engine
        self._type = type
        self._width = width
        self._operand_width = width if isinstance(width, int) else None
        self._data_type = data_type
        self._is_synchronized = is_synchronized
        self._pulsing = pulsing
        self._assignment = assignment

    @property
    def engine(self) -> EngineProto:
        """NoRTL engine for this signal."""
        return self._engine

    @property
    def type(self) -> SIGNAL_TYPES:
        """Signal type."""
        return self._type

    @property
    def pulsing(self) -> bool:
        """Shows, if the signal is self-resetting to zero after one cycle."""
        return self._pulsing

    @property
    def assignment(self) -> Optional[Renderable]:
        """Source expression for combination assignment."""
        return self._assignment

    @property
    def escaped_name(self) -> str:
        """Name of signal with any special characters escaped.

        For regular signals, this is equal to their name. For slice signals, it contains the position.
        """
        return self.name

    @property
    def width(self) -> Union[int, ParameterProto, Renderable]:
        """Signal width in bits."""
        return self._width

    @property
    def operand_width(self) -> Optional[int]:
        """Indicates the width when used as an operand.

        A width of None means that the width is not fixed during execution of noRTL.
        This is the case, if the signal width is based on a parameter.
        """
        return self._operand_width

    @property
    def data_type(self) -> str:
        """Data type of the signal (e.g., 'logic', 'reg', etc.)."""
        return self._data_type

    def render(self, target: Optional[str] = None) -> str:
        """Render value to target language.

        Arguments:
            target: Target language.
        """
        return self.name

    def __getitem__(self, index: Union[int, IntSlice]) -> 'SignalSlice':
        return SignalSlice(self, index)

    def overlaps_with(self, other: AssignmentTarget) -> Union[bool, Literal['partial']]:
        """Check if signal overlaps with other signal or signal slice."""
        return self.name == other.name

    def read_access(self, ignore: Set[ACCESS_CHECKS] = set()) -> None:
        """Register read access from the current thread.

        If the current thread differs from the last write acess, will invoke the access checker.

        Raises:
            ExclusiveReadError: If the signal was read from more than one thread.
            NonIdenticalRWError: If the signals was written by one, and read from another thread.
        """
        if self.assignment is not None:
            # Trigger read access on the assignment expression
            self.assignment.read_access(ignore=ignore)
        else:
            super().read_access(ignore=ignore)

    def write_access(self, ignore: Set[ACCESS_CHECKS] = set()) -> None:
        """Register write access from the current thread.

        If the current thread differs from the last write acess, will invoke the access checker.

        Raises:
            ExclusiveWriteError: If the signals was written by more than one thread.
            NonIdenticalRWError: If the signals was written by one, and read from another thread.
            WriteViolationError: If the signal is read-only.
        """
        if self.type == 'input':
            raise WriteViolationError(f'Input signal {self.name} is read-only.')
        if self.assignment is not None:
            raise WriteViolationError(f'Signal {self.name} is assigned to the expression {self.assignment}. It is read-only.')
        else:
            super().write_access(ignore=ignore)


class _BaseSlice(_BaseSignal):
    """Intermediate class for signal slices."""

    def __init__(self, signal: SignalProto, index: Union[int, IntSlice]) -> None:
        super().__init__(signal.name)

        self._base_signal = signal
        self._index = index

        if isinstance(signal.width, int):
            if signal.width <= 1:
                raise IndexError(f'Unable to slice signal with width {signal.width}!')

        if isinstance(index, int):
            if index < 0:
                raise IndexError(f'Index {index} is out of bounds for signal {signal.name} with width {signal.width}')
            if isinstance(signal.width, int) and index not in range(0, signal.width):
                raise IndexError(f'Index {index} is out of bounds for signal {signal.name} with width {signal.width}')

            self._width = 1
            self._bitorder: Optional[BIT_ORDER] = None
        else:
            start, stop, bitorder = validate_slice(index)
            self._width = stop - start + 1
            self._bitorder = bitorder

    @property
    def base_signal(self) -> SignalProto:
        """Full-width signal of this slice."""
        return self._base_signal

    @property
    def index(self) -> Union[int, IntSlice]:
        """Index of the signal."""
        return self._index

    # Properties forwarded to full-width signal
    @property
    def engine(self) -> EngineProto:
        """Finite state machine."""
        return self.base_signal.engine

    @property
    def type(self) -> SIGNAL_TYPES:
        """Signal type."""
        return self.base_signal.type

    @property
    def pulsing(self) -> bool:
        """Shows, if the signal is self-resetting to zero after one cycle."""
        return self.base_signal.pulsing

    @property
    def escaped_name(self) -> str:
        """Name of signal with any special characters escaped.

        For regular signals, this is equal to their name. For slice signals, it contains the position.
        """
        if isinstance(self.index, int):
            return f'{self.name}_{self.index}'
        else:
            return f'{self.name}_{self.index.start}to{self.index.stop}'

    @property
    def data_type(self) -> str:
        """Data type of the signal (e.g., 'logic', 'reg', etc.)."""
        return self.base_signal.data_type

    @property
    def _is_synchronized(self) -> bool:
        """Indicates if the this signal is synchronized to the local clock domain."""
        return self.base_signal._is_synchronized

    # Additional properties
    @property
    def width(self) -> int:
        """Signal width in bits."""
        return self._width

    @property
    def bitorder(self) -> Optional[BIT_ORDER]:
        """Bit order of signal."""
        return self._bitorder

    @property
    def operand_width(self) -> int:
        """Indicates the width when used as an operand.

        A width of None means that the width is not fixed during execution of noRTL.
        This is the case, if the signal width is based on a parameter.
        """
        return self.width

    def overlaps_with(self, other: AssignmentTarget) -> Union[bool, Literal['partial']]:
        """Check if signal slice overlaps with other signal or signal slice."""

        if self.name != other.name:
            return False

        # Unwrap content of modifier
        if isinstance(other, BaseModifier):
            other = other.content

        if isinstance(other, _BaseSlice):
            # Signal slice, check if it overlaps
            if isinstance(self.index, int) and isinstance(other.index, int):
                # Full overlap or none
                return self.index == other.index
            elif isinstance(self.index, slice) and isinstance(other.index, slice) and self.index == other.index:
                # Full overlap, if the slices are exactly the same
                return True

            own_indexes = set(list_indexes(self.index))
            other_indexes = set(list_indexes(other.index))

            if own_indexes.isdisjoint(other_indexes):
                return False
            if own_indexes == other_indexes:
                return True

        # In all other cases (overlap of signal and slice, parametric width, partial overlap), treat overlap as partial
        return 'partial'

    def render(self, target: Optional[str] = None) -> str:
        """Render value to target language.

        Arguments:
            target: Target language.
        """
        if isinstance(self.index, int):
            return f'{self.name}[{self.index}]'
        elif self.index.start == self.index.stop:
            return f'{self.name}[{self.index.start}]'
        else:
            return f'{self.name}[{self.index.start}:{self.index.stop}]'

    def __getitem__(self, index: Union[int, IntSlice]) -> Self:
        if isinstance(self.index, int):
            if index == 0:
                return self  # Allow indexing [0] of a single bit slice again
            else:
                raise IndexError(f'Unable to slice {index} from single-bit signal slice {self.escaped_name}: Only index 0 can be sliced.')
        else:
            # Assemble list of own indexes, with reference to the base signal
            own_indexes = list_indexes(self.index)
            own_start: int = own_indexes[0]
            own_stop: int = own_indexes[-1]

            # Check that the new index doesn't go out of range (this is not caught by pick_indexes)
            if isinstance(index, int):
                if index < 0 or index > (own_stop - own_start):
                    raise IndexError(f'Unable to slice {index} from signal slice {self.escaped_name}: Index is out of range.')
            else:
                start, stop, bitorder = validate_slice(index)
                if start < 0 or stop > (own_stop - own_start):
                    raise IndexError(f'Unable to slice {index} from signal slice {self.escaped_name}: Index is out of range.')
                if bitorder != self.bitorder:
                    raise IndexError(f'Unable to slice {index} from signal slice {self.escaped_name}: Reversing the bit order is not allowed')

            # Pick the indexes for the nested slice, with reference to the base signal
            new_indexes = pick_indexes(own_indexes, index)

            if len(new_indexes) == 0:
                raise IndexError(f'Unable to slice {index} from signal slice {self.escaped_name}: slice result in zero length signal.')
            if len(new_indexes) == 1:
                return type(self)(self.base_signal, new_indexes[0])

            # As we don't support multiple indexes, we only need to find the new minimum and maximum indexes
            if bitorder == 'H:L':
                return type(self)(self.base_signal, slice(max(new_indexes), min(new_indexes)))
            else:
                return type(self)(self.base_signal, slice(min(new_indexes), max(new_indexes)))


class SignalSlice(_BaseSlice, _EventSourceSignal[SignalSliceProto]):
    """Slice of a signal."""

    def __init__(self, signal: SignalProto, index: Union[int, IntSlice]) -> None:
        # if not isinstance(signal.width, int):
        # TODO: Do we need to pass metadata along?
        # TODO: parameters or renderable widths cannot be validated
        # raise NotImplementedError('Only signals with discrete width can be sliced!')  # noqa: ERA001
        super().__init__(signal, index)

    # Forward access control to base signal
    @property
    def access_checker(self) -> StaticAccessCheckerProto:
        """Static access checker."""
        return self.base_signal.access_checker

    @property
    def read_accesses(self) -> Set[StaticAccessProto]:
        """Mutable set of read accesses to this signal."""
        return self.base_signal.read_accesses

    @property
    def write_accesses(self) -> Set[StaticAccessProto]:
        """Mutable set of write accesses to this signal."""
        return self.base_signal.write_accesses

    @property
    def last_read_access_thread(self) -> Optional[ThreadProto]:
        """Thread performing the last read access."""
        return self.base_signal.last_read_access_thread

    @last_read_access_thread.setter
    def last_read_access_thread(self, value: ThreadProto) -> None:
        """Thread performing the last read access."""
        self.base_signal.last_read_access_thread = value

    @property
    def last_write_access_thread(self) -> Optional[ThreadProto]:
        """Thread performing the last write access."""
        return self.base_signal.last_write_access_thread

    @last_write_access_thread.setter
    def last_write_access_thread(self, value: ThreadProto) -> None:
        """Thread performing the last write access."""
        self.base_signal.last_write_access_thread = value

    def as_scratch_signal(self, zone: Optional[MemoryZoneProto] = None) -> 'ScratchSignal':
        """Turn SignalSlice into ScratchSignal, owned by the current thread."""
        if zone is None:
            zone = self.engine.scratch_manager.active_zone
        return ScratchSignal(self.base_signal, self.index, zone)


class ScratchSignal(_BaseSlice, _AccessControlledSignal):
    """A scratch signal is a special kind of signal slice, that is only valid for a limited time."""

    def __init__(self, signal: SignalProto, index: Union[int, IntSlice], zone: Optional[MemoryZoneProto] = None) -> None:
        super().__init__(signal, index)

        # Register scratch signal
        # This is done here, to correctly register slices of scratch registers
        if zone is None:
            zone = self.engine.scratch_manager.active_zone
        zone.active_view._register_signal(self)

        # Access control for scratch signals
        self._owner = signal.engine.current_thread
        self._released = False
        self._context_ctr = 0
        self._context_ctr_active = True
        self._zone = zone

        # Remember, where the scratch signal was created
        self.creator_frames = self.engine.tracer.current_trace

    @property
    def owner(self) -> ThreadProto:
        """Owner of this scratch signal."""
        return self._owner

    @property
    def zone(self) -> MemoryZoneProto:
        """Memory zone for this scratch signal."""
        return self._zone

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]) -> None:
        if exc_type is not None:
            return
        self.release()

    def enter_context(self) -> None:
        """Context counting method.

        The claim/release logic relies on the concept, that a scratch variable is to be released in the context where it has been created.
        A context is represented by a Condition or a Loop construct. Both are created with context managers and may be nested.
        The idea of this function is to count the 'depth' of context nests that we are currently working with.

        In this way, we can detect, if the user tries to release the signal in a different context than creation. Once the release function is called,
        we stop context counting since the scratch signal is now inactive and the claim/release-control is now realized based on the currently running threads.

        This concept assumes, that each context manager triggers the `enter_context` function during enter and the `exit_context` function during exit.
        The `exit_context` function automatically releases the signal, if we leave the claiming context.
        """
        if self._context_ctr_active:
            self._context_ctr += 1

    def exit_context(self) -> None:
        """Context counting method. Explanation in exit_context function."""
        if self._context_ctr_active:
            self._context_ctr -= 1

            if self._context_ctr == -1 and self.owner == self.engine.current_thread:
                self._context_ctr = 0
                self.release()

    # Access control
    def write_access(self, ignore: Set[ACCESS_CHECKS] = set()) -> None:
        """Register write access from the current thread.

        If the current thread differs from the last write acess, will invoke the access checker.

        Raises:
            ExclusiveWriteError: If the signals was written by more than one thread.
            NonIdenticalRWError: If the signals was written by one, and read from another thread.
            AccessAfterReleaseError: If the scratch signal was released.
        """
        if self.released:
            raise AccessAfterReleaseError(f'Signal {self} was accessed after being released.')
        super().write_access(ignore=ignore)

    def read_access(self, ignore: Set[ACCESS_CHECKS] = set()) -> None:
        """Register read access from the current thread.

        If the current thread differs from the last write acess, will invoke the access checker.

        Raises:
            ExclusiveReadError: If the signal was read from more than one thread.
            NonIdenticalRWError: If the signals was written by one, and read from another thread.
            AccessAfterReleaseError: If the scratch signal was released.
        """
        if self.released:
            raise AccessAfterReleaseError('Tried to read from a signal that has been released previously!')
        super().read_access(ignore=ignore)

    @property
    def released(self) -> bool:
        """Findout, if a signal has been release yet.

        The claim/release control works based on two principles:

        1. A scratch signal may only be claimed and released in a single code block. Example:
        ```python
        f = CoreEngine("my_engine")

        with Condition(f, some_condition):
            s = # new scratch signal

            with ForLoop(...):
                # s may not be released here

            s.release() # s can be released here, since it is the same context

        # After the context has ended, s is automatically released.
        ```

        2. A scratch signal will appear as non-released in parallel threads and will be released once the owner thread ends. Additional access control applies.
        ```python
        f = CoreEngine("my_engine")

        with Fork(f, "my_fork") as f1:
            s = # new scratch_signals
            #...
            s.release()
        with Fork(f, "my_second_fork") as f2:
            assert s.released == False # Parallel running thread the scratch pad location!

        f1.wait_for_finish()

        # s is released automatically once the thread has finished.
        ```

        """
        if self.owner != self.base_signal.engine.current_thread:
            return not self.owner.running
        return self._released

    def reclaim(self) -> None:
        """Reclaim the scratch signal.

        This is dangerous and meant to be used for internal purposes.

        Even if the ressource can be reclaimed without collisions, the data within may no longer be valid. This entirely depends on where (in whichs state) the signal was reclaimed.
        """
        if not self.released:
            raise RuntimeError('Scratch signal was not released before reclaim. This indicates internal issues.')
        if (active_zone := self.engine.scratch_manager.active_zone) is not self.zone:
            raise RuntimeError(f'Scratch signal belongs to memory zone {self.zone.name} and cannot be reclaimed from zone {active_zone.name}.')
        if self not in self.zone.scratch_signals:
            raise RuntimeError(
                f'Scratch signal does not belong to the active view of memory zone {self.zone.name} and can therefore not be reclaimed.'
            )

        self._owner = self.engine.current_thread
        self._context_ctr = 0
        self._context_ctr_active = True
        self._released = False

    def release(self, force: bool = False) -> None:
        """Release the scratch signal.

        After the signal has been released, it's allocated space is freed up and can be re-used by other signals.
        """
        if self.owner != self.base_signal.engine.current_thread and not force:
            raise ValueError('Scratch signal may only be released in owning thread!')
        if self._context_ctr != 0 and not force:
            raise ValueError('Scratch signals need to be released in the context where they were created!')

        self._context_ctr_active = False
        self._released = True

    @property
    def _active_states(self) -> Set[str]:
        if self.has_metadata('cache_active_state_enabled'):
            if self.get_metadata('cache_active_state_enabled') and self.has_metadata('cache_active_state_data'):
                return self.get_metadata('cache_active_state_data')  # type: ignore

        ret = set()
        for statelst in self.engine.states.values():
            for state in statelst:
                for s in state.active_scratch_signals:
                    if self is s:  # FIXME: Weird containment, scratch_signal in self.active_scratch signals is always True?
                        ret.add(state.name)

        if self.has_metadata('cache_active_state_enabled'):
            if self.get_metadata('cache_active_state_enabled'):
                self.set_metadata('cache_active_state_data', ret)
        return ret

    def states_disjoint(self, other: Self) -> bool:
        """This function calculates if the current scratch signal and the other scratch signal are never active (i.a. non-released) in the same states."""

        other_active_states = other._active_states
        self_active_states = self._active_states

        return self_active_states.isdisjoint(other_active_states)

    def call_stack_similarity(self, other: Self) -> int:
        own_stack = self.creator_frames
        other_stack = other.creator_frames

        def frameinfo(frame: FrameInfo) -> str:
            return f'{frame.filename}, {frame.lineno}: {frame.function} {frame.code_context}\n'

        if frameinfo(own_stack[-1]) != frameinfo(other_stack[-1]):
            return 1
        else:
            return -1

    def _creator_frameinfo(self) -> str:
        frame = self.creator_frames[-1]
        return f'{frame.filename}, {frame.lineno}: {frame.function} {frame.code_context}\n'



"""Logic operations."""

from typing import Optional, TypeVar, Union

from nortl.core.protocols import Operand, Renderable
from nortl.core.renderers.operations.base import TwoSideRenderer, const_unpack

__all__ = [
    'And',
    'ExclusiveOr',
    'LeftShift',
    'Or',
    'RightShift',
]

T_Left = TypeVar('T_Left', bound=Renderable)
T_Right = TypeVar('T_Right', bound=Renderable)


class And(TwoSideRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        return f'({self.left} & {self.right})'

    @staticmethod
    def eval(left: Operand, right: Operand) -> Optional[int]:
        """Evaluate operation into constant value or single operand.

        Short circuits for x & 0.
        """
        left_ = const_unpack(left)
        right_ = const_unpack(right)

        # Short circuiting: x & 0 = 0
        if left_ == 0 or right_ == 0:
            return 0
        # FIXME implement short circuiting for: x & 1 (must respect operand widths)

        if left_ is not None and right_ is not None:
            return left_ & right_
        return None


class Or(TwoSideRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        return f'({self.left} | {self.right})'

    @staticmethod
    def eval(left: Union[T_Left, int, bool], right: Union[T_Right, int, bool]) -> Optional[Union[T_Left, T_Right, int]]:
        """Evaluate operation into constant value.

        Short circuits for x | 0.
        """
        left_ = const_unpack(left)
        right_ = const_unpack(right)

        # Short circuiting: x | 0 = x
        if left_ == 0:
            return right
        if right_ == 0:
            return left
        # FIXME implement short circuiting for: x | 1 (must respect operand widths)

        if left_ is not None and right_ is not None:
            return left_ | right_
        return None


class ExclusiveOr(TwoSideRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        return f'({self.left} ^ {self.right})'

    @staticmethod
    def eval(left: Union[T_Left, int, bool], right: Union[T_Right, int, bool]) -> Optional[Union[T_Left, T_Right, int]]:
        """Evaluate operation into constant value.

        Short circuits for x ^ 0.
        """
        left_ = const_unpack(left)
        right_ = const_unpack(right)

        # Short circuiting: x ^ 0 = x
        if left_ == 0:
            return right
        if right_ == 0:
            return left

        if left_ is not None and right_ is not None:
            return left_ ^ right_
        return None


class LeftShift(TwoSideRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        return f'({self.left} << {self.right})'

    @staticmethod
    def eval(left: Union[T_Left, int, bool], right: Union[T_Right, int, bool]) -> Optional[Union[T_Left, T_Right, int]]:
        """Evaluate operation into constant value.

        Short circuits for shift of/by 0.
        """
        left_ = const_unpack(left)
        right_ = const_unpack(right)

        # Short circuiting: x << 0 = x and 0 << x = 0
        if left_ == 0:
            return 0
        if right_ == 0:
            return left

        if left_ is not None and right_ is not None:
            return left_ << right_
        return None


class RightShift(TwoSideRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        return f'({self.left} >> {self.right})'

    @staticmethod
    def eval(left: Union[T_Left, int, bool], right: Union[T_Right, int, bool]) -> Optional[Union[T_Left, T_Right, int]]:
        """Evaluate operation into constant value.

        Short circuits for shift of/by 0.
        """
        left_ = const_unpack(left)
        right_ = const_unpack(right)

        # Short circuiting: x >> 0 = x and 0 >> x = 0
        if left_ == 0:
            return 0
        if right_ == 0:
            return left

        if left_ is not None and right_ is not None:
            return left_ >> right_
        return None



"""Tests correct handling of input and output slots, as well as signal-expiration."""

import re
from pathlib import Path
from typing import Protocol, Union

from nortl import Engine, Segment
from nortl.core.protocols import EngineProto, ScratchSignalProto

ARTIFACTS_PATH = Path(__file__).parent / '..' / 'artifacts'


class TestExecutor(Protocol):
    __test__: bool = False

    def __call__(self, engine: Engine, clock_gating: bool = False) -> str: ...


@Segment.with_input_slots(sig=8).with_output_slots(8)
def a(engine: EngineProto, sig: Union[int, ScratchSignalProto]) -> ScratchSignalProto:
    res = engine.define_scratch(8)
    engine.set(res, sig + 1)
    return res


@Segment.with_input_slots(sig=8).with_output_slots(8)
def b(engine: EngineProto, sig: Union[int, ScratchSignalProto]) -> ScratchSignalProto:
    res = a(engine, sig)
    return res


@Segment.with_input_slots(sig=8).with_output_slots(8)
def c(engine: EngineProto, sig: Union[int, ScratchSignalProto]) -> ScratchSignalProto:
    res = a(engine, sig)
    return res


def test_call_return_address(execute_test: TestExecutor) -> None:
    """Test correct operation of the return address for segments."""
    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    output = engine.define_output('OUT', 8, reset_value=0)
    count = engine.define_local('count', 2, reset_value=0)
    engine.sync()

    # Manually create a for-loop that causes the segment to break, if the return addresses don't work
    engine.set(count, count + 1)
    assert len(engine.states['main']) == 2  # (IDLE + STATE_1)

    start_state = engine.current_state
    x = a(engine, 0)
    engine.set(output, x)  # Should be 1  == STATE_3
    assert len(engine.states['main']) == 4  # + Call, Return for a()

    x = b(engine, 8)
    engine.set(output, x + 1)  # Should be 10
    assert len(engine.states['main']) == 7  # + Call, Return for b(); Call for a()

    x = a(engine, 0)
    engine.set(output, x + 2)  # Should be 3
    assert len(engine.states['main']) == 8  # + Call for a()

    x = c(engine, 4)
    engine.set(output, x + 3)  # Should be 8
    assert len(engine.states['main']) == 11  # + Call, Return for c(); Call for a() + b()

    # End for loop
    engine.jump_if(count < 2, start_state, engine.next_state)

    # Execute test
    result = execute_test(engine)
    print(result)
    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = ['OUT=0'] + ['OUT=1', 'OUT=10', 'OUT=3', 'OUT=8'] * 2

    assert res == expected



import shutil
from pathlib import Path
from subprocess import PIPE, run
from tempfile import TemporaryDirectory
from typing import Protocol

import pytest

from nortl import Engine
from nortl.renderer.verilog_renderer import VerilogRenderer


class SimulatorError(Exception):
    pass


class TestExecutor(Protocol):
    def __call__(self, engine: Engine, clock_gating: bool = False) -> str: ...


@pytest.fixture
def execute_test() -> TestExecutor:
    def execute_test(engine: Engine, clock_gating: bool = False) -> str:
        """Helper function that uses iverilog to simulate the testbench with the engine.

        Args:
            engine (Engine): Engine to be tested
            clock_gating (bool): If clock gating should be used in this test
        """

        verilog_file = Path(__file__).parent.parent / 'verilog' / 'testbench.sv'

        with TemporaryDirectory() as tempdir_str:
            tempdir = Path(tempdir_str)

            # Copy verilog testbench to tempdir
            shutil.copy(verilog_file, tempdir)

            # Render engine to verilog
            renderer = VerilogRenderer(engine, clock_gating=clock_gating)
            with open(tempdir / 'dut.sv', 'w') as file:
                file.write(renderer.render())

            # Save code as artifact
            with open((Path(__file__).parent.parent / 'artifacts') / 'dut.sv', 'w') as file:
                file.write(renderer.render())

            # Execute icarus verilog
            command = ['iverilog', '-g2005-sv', 'dut.sv', 'testbench.sv']
            result = run(command, cwd=tempdir, stdout=PIPE, stderr=PIPE, universal_newlines=True)

            if result.returncode != 0:
                raise SimulatorError(f'Icarus Verilog Compiler Error: \n {result.stdout} \n {result.stderr}')

            # run simulation
            result = run(['vvp', 'a.out'], cwd=tempdir, stdout=PIPE, stderr=PIPE, universal_newlines=True)

            shutil.copyfile(tempdir / 'out.vcd', (Path(__file__).parent.parent / 'artifacts') / 'out.vcd')

            if result.returncode != 0:
                raise SimulatorError(f'Icarus Verilog Runner Error: \n {result.stdout} \n {result.stderr}')

            return result.stdout

    return execute_test
